import itertools
from typing import Dict, List, Tuple

import numpy as np
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_CYDIR,
    INSTR_CZ,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.models.qerrormodels import (
    DepolarNoiseModel,
    QuantumErrorModel,
    T1T2NoiseModel,
)
from netsquid.components.qprocessor import PhysicalInstruction, QuantumProcessor
from netsquid.qubits.operators import Operator
from netsquid_magic.link_layer import (
    MagicLinkLayerProtocolWithSignaling,
    SingleClickTranslationUnit,
)
from netsquid_magic.magic_distributor import (
    DepolariseWithFailureMagicDistributor,
    DoubleClickMagicDistributor,
    PerfectStateMagicDistributor,
)
from netsquid_physlayer.heralded_connection import MiddleHeraldedConnection

from qoala.lang.ehi import EhiLinkInfo, EhiNetworkInfo

# Ignore type since whole 'config' module is ignored by mypy
from qoala.runtime.config import (  # type: ignore
    DepolariseOldLinkConfig,
    GenericQDeviceConfig,
    HeraldedOldLinkConfig,
    LinkConfig,
    NVQDeviceConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
)
from qoala.runtime.environment import NetworkInfo
from qoala.runtime.lhi import (
    INSTR_MEASURE_INSTANT,
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiProcNodeInfo,
    LhiTopology,
    LhiTopologyBuilder,
)
from qoala.runtime.lhi_nv_compat import LhiTopologyBuilderForOldNV
from qoala.runtime.lhi_to_ehi import (
    GenericToVanillaInterface,
    LhiConverter,
    NativeToFlavourInterface,
    NvToNvInterface,
)
from qoala.sim.entdist.entdist import EntDist
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.network import ProcNodeNetwork
from qoala.sim.procnode import ProcNode
from qoala.util.math import fidelity_to_prob_max_mixed


def build_qprocessor_from_topology(
    name: str, topology: LhiTopology
) -> QuantumProcessor:
    num_qubits = len(topology.qubit_infos)

    mem_noise_models: List[QuantumErrorModel] = []
    for i in range(num_qubits):
        info = topology.qubit_infos[i]
        noise_model = info.error_model(**info.error_model_kwargs)
        mem_noise_models.append(noise_model)

    phys_instructions: List[PhysicalInstruction] = []
    # single-qubit gates
    for qubit_id, gate_infos in topology.single_gate_infos.items():
        for gate_info in gate_infos:
            # TODO: refactor this hack
            if gate_info.instruction == INSTR_MEASURE_INSTANT:
                duration = 0.0
            else:
                duration = gate_info.duration

            phys_instr = PhysicalInstruction(
                instruction=gate_info.instruction,
                duration=duration,
                topology=[qubit_id],
                quantum_noise_model=gate_info.error_model(
                    **gate_info.error_model_kwargs
                ),
            )
            phys_instructions.append(phys_instr)

    # multi-qubit gates
    for multi_qubit, gate_infos in topology.multi_gate_infos.items():
        qubit_ids = tuple(multi_qubit.qubit_ids)
        for gate_info in gate_infos:
            phys_instr = PhysicalInstruction(
                instruction=gate_info.instruction,
                duration=gate_info.duration,
                topology=[qubit_ids],
                quantum_noise_model=gate_info.error_model(
                    **gate_info.error_model_kwargs
                ),
            )
            phys_instructions.append(phys_instr)

    return QuantumProcessor(
        name=name,
        num_positions=num_qubits,
        mem_noise_models=mem_noise_models,
        phys_instructions=phys_instructions,
    )


def build_generic_qprocessor(name: str, cfg: GenericQDeviceConfig) -> QuantumProcessor:
    phys_instructions = []

    single_qubit_gate_noise = DepolarNoiseModel(
        depolar_rate=cfg.single_qubit_gate_depolar_prob, time_independent=True
    )

    two_qubit_gate_noise = DepolarNoiseModel(
        depolar_rate=cfg.two_qubit_gate_depolar_prob, time_independent=True
    )

    phys_instructions.append(
        PhysicalInstruction(
            INSTR_INIT,
            duration=cfg.init_time,
        )
    )

    for instr in [
        INSTR_ROT_X,
        INSTR_ROT_Y,
        INSTR_ROT_Z,
        INSTR_X,
        INSTR_Y,
        INSTR_Z,
        INSTR_H,
    ]:
        phys_instructions.append(
            PhysicalInstruction(
                instr,
                quantum_noise_model=single_qubit_gate_noise,
                duration=cfg.single_qubit_gate_time,
            )
        )

    for instr in [INSTR_CNOT, INSTR_CZ]:
        phys_instructions.append(
            PhysicalInstruction(
                instr,
                quantum_noise_model=two_qubit_gate_noise,
                duration=cfg.two_qubit_gate_time,
            )
        )

    phys_instr_measure = PhysicalInstruction(
        INSTR_MEASURE,
        duration=cfg.measure_time,
    )
    phys_instructions.append(phys_instr_measure)

    electron_qubit_noise = T1T2NoiseModel(T1=cfg.T1, T2=cfg.T2)
    mem_noise_models = [electron_qubit_noise] * cfg.num_qubits
    qmem = QuantumProcessor(
        name=name,
        num_positions=cfg.num_qubits,
        mem_noise_models=mem_noise_models,
        phys_instructions=phys_instructions,
    )
    return qmem


def build_nv_qprocessor(name: str, cfg: NVQDeviceConfig) -> QuantumProcessor:
    # noise models for single- and multi-qubit operations
    electron_init_noise = DepolarNoiseModel(
        depolar_rate=cfg.electron_init_depolar_prob, time_independent=True
    )

    electron_single_qubit_noise = DepolarNoiseModel(
        depolar_rate=cfg.electron_single_qubit_depolar_prob, time_independent=True
    )

    carbon_init_noise = DepolarNoiseModel(
        depolar_rate=cfg.carbon_init_depolar_prob, time_independent=True
    )

    carbon_z_rot_noise = DepolarNoiseModel(
        depolar_rate=cfg.carbon_z_rot_depolar_prob, time_independent=True
    )

    ec_noise = DepolarNoiseModel(
        depolar_rate=cfg.ec_gate_depolar_prob, time_independent=True
    )

    electron_qubit_noise = T1T2NoiseModel(T1=cfg.electron_T1, T2=cfg.electron_T2)

    carbon_qubit_noise = T1T2NoiseModel(T1=cfg.carbon_T1, T2=cfg.carbon_T2)

    # defining gates and their gate times

    phys_instructions = []

    electron_position = 0
    carbon_positions = [pos + 1 for pos in range(cfg.num_qubits - 1)]

    phys_instructions.append(
        PhysicalInstruction(
            INSTR_INIT,
            topology=carbon_positions,
            quantum_noise_model=carbon_init_noise,
            duration=cfg.carbon_init,
        )
    )

    for instr, dur in zip(
        [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z],
        [cfg.carbon_rot_x, cfg.carbon_rot_y, cfg.carbon_rot_z],
    ):
        phys_instructions.append(
            PhysicalInstruction(
                instr,
                topology=carbon_positions,
                quantum_noise_model=carbon_z_rot_noise,
                duration=dur,
            )
        )

    phys_instructions.append(
        PhysicalInstruction(
            INSTR_INIT,
            topology=[electron_position],
            quantum_noise_model=electron_init_noise,
            duration=cfg.electron_init,
        )
    )

    for instr, dur in zip(
        [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z],
        [cfg.electron_rot_x, cfg.electron_rot_y, cfg.electron_rot_z],
    ):
        phys_instructions.append(
            PhysicalInstruction(
                instr,
                topology=[electron_position],
                quantum_noise_model=electron_single_qubit_noise,
                duration=dur,
            )
        )

    electron_carbon_topologies = [
        (electron_position, carbon_pos) for carbon_pos in carbon_positions
    ]
    phys_instructions.append(
        PhysicalInstruction(
            INSTR_CXDIR,
            topology=electron_carbon_topologies,
            quantum_noise_model=ec_noise,
            duration=cfg.ec_controlled_dir_x,
        )
    )

    phys_instructions.append(
        PhysicalInstruction(
            INSTR_CYDIR,
            topology=electron_carbon_topologies,
            quantum_noise_model=ec_noise,
            duration=cfg.ec_controlled_dir_y,
        )
    )

    M0 = Operator(
        "M0", np.diag([np.sqrt(1 - cfg.prob_error_0), np.sqrt(cfg.prob_error_1)])
    )
    M1 = Operator(
        "M1", np.diag([np.sqrt(cfg.prob_error_0), np.sqrt(1 - cfg.prob_error_1)])
    )

    # hack to set imperfect measurements
    INSTR_MEASURE._meas_operators = [M0, M1]

    phys_instr_measure = PhysicalInstruction(
        INSTR_MEASURE,
        topology=[electron_position],
        quantum_noise_model=None,
        duration=cfg.measure,
    )

    phys_instructions.append(phys_instr_measure)

    # add qubits
    mem_noise_models = [electron_qubit_noise] + [carbon_qubit_noise] * len(
        carbon_positions
    )
    qmem = QuantumProcessor(
        name=name,
        num_positions=cfg.num_qubits,
        mem_noise_models=mem_noise_models,
        phys_instructions=phys_instructions,
    )
    return qmem


def build_procnode(
    cfg: ProcNodeConfig, network_info: NetworkInfo, network_ehi: EhiNetworkInfo
) -> ProcNode:
    # TODO: Refactor ad-hoc way of old NV config
    # TODO: Refactor how ntf interface is configured!
    ntf_interface: NativeToFlavourInterface
    if cfg.topology is not None:
        topology = LhiTopologyBuilder.from_config(cfg.topology)
        ntf_interface = GenericToVanillaInterface()
    if cfg.nv_config is not None:
        topology = LhiTopologyBuilderForOldNV.from_nv_config(cfg.nv_config)
        ntf_interface = NvToNvInterface()

    qprocessor = build_qprocessor_from_topology(name=cfg.node_name, topology=topology)
    latencies = LhiLatencies.from_config(cfg.latencies)
    procnode = ProcNode(
        cfg.node_name,
        network_info=network_info,
        qprocessor=qprocessor,
        qdevice_topology=topology,
        latencies=latencies,
        ntf_interface=ntf_interface,
        node_id=cfg.node_id,
        network_ehi=network_ehi,
    )

    # TODO: refactor this hack
    procnode.qnos.processor._latencies.qnos_instr_time = cfg.latencies.qnos_instr_time
    procnode.host.processor._latencies.host_instr_time = cfg.latencies.host_instr_time
    procnode.host.processor._latencies.host_peer_latency = (
        cfg.latencies.host_peer_latency
    )
    return procnode


def build_ll_protocol(
    config: LinkConfig, proc_node1: ProcNode, proc_node2: ProcNode
) -> MagicLinkLayerProtocolWithSignaling:
    if config.typ == "perfect":
        link_dist = PerfectStateMagicDistributor(
            nodes=[proc_node1.node, proc_node2.node], state_delay=0
        )
    elif config.typ == "depolarise":
        link_cfg = config.cfg
        if not isinstance(link_cfg, DepolariseOldLinkConfig):
            link_cfg = DepolariseOldLinkConfig(**config.cfg)
        prob_max_mixed = fidelity_to_prob_max_mixed(link_cfg.fidelity)
        link_dist = DepolariseWithFailureMagicDistributor(
            nodes=[proc_node1.node, proc_node2.node],
            prob_max_mixed=prob_max_mixed,
            prob_success=link_cfg.prob_success,
            t_cycle=link_cfg.t_cycle,
        )
    # TODO: decide whether this is still wanted/needed
    # elif config.typ == "nv":
    #     link_cfg = config.cfg
    #     if not isinstance(link_cfg, NVLinkConfig):
    #         link_cfg = NVLinkConfig(**config.cfg)
    #     link_dist = NVSingleClickMagicDistributor(
    #         nodes=[proc_node1.node, proc_node2.node],
    #         length_A=link_cfg.length_A,
    #         length_B=link_cfg.length_B,
    #         full_cycle=link_cfg.full_cycle,
    #         cycle_time=link_cfg.cycle_time,
    #         alpha=link_cfg.alpha,
    #     )
    elif config.typ == "heralded":
        link_cfg = config.cfg
        if not isinstance(link_cfg, HeraldedOldLinkConfig):
            link_cfg = HeraldedOldLinkConfig(**config.cfg)
        connection = MiddleHeraldedConnection(name="heralded_conn", **link_cfg.dict())
        link_dist = DoubleClickMagicDistributor(
            [proc_node1.node, proc_node2.node], connection
        )
    else:
        raise ValueError

    return MagicLinkLayerProtocolWithSignaling(
        nodes=[proc_node1.node, proc_node2.node],
        magic_distributor=link_dist,
        translation_unit=SingleClickTranslationUnit(),
    )


def build_network(
    config: ProcNodeNetworkConfig,
    network_info: NetworkInfo,
) -> ProcNodeNetwork:
    procnodes: Dict[str, ProcNode] = {}

    ehi_links: Dict[Tuple[int, int], EhiLinkInfo] = {}
    for link_between_nodes in config.links:
        lhi_link = LhiLinkInfo.from_config(link_between_nodes.link_config)
        ehi_link = LhiConverter.link_info_to_ehi(lhi_link)
        ids = (link_between_nodes.node_id1, link_between_nodes.node_id2)
        ehi_links[ids] = ehi_link
    network_ehi = EhiNetworkInfo(ehi_links)

    for cfg in config.nodes:
        procnodes[cfg.node_name] = build_procnode(cfg, network_info, network_ehi)

    ns_nodes = [procnode.node for procnode in procnodes.values()]
    entdistcomp = EntDistComponent(network_info)
    entdist = EntDist(nodes=ns_nodes, network_info=network_info, comp=entdistcomp)

    for link_between_nodes in config.links:
        link = LhiLinkInfo.from_config(link_between_nodes.link_config)
        n1 = link_between_nodes.node_id1
        n2 = link_between_nodes.node_id2
        entdist.add_sampler(n1, n2, link)

    for (_, s1), (_, s2) in itertools.combinations(procnodes.items(), 2):
        s1.connect_to(s2)

    for name, procnode in procnodes.items():
        procnode.node.entdist_out_port.connect(entdistcomp.node_in_port(name))
        procnode.node.entdist_in_port.connect(entdistcomp.node_out_port(name))

    return ProcNodeNetwork(procnodes, entdist)


def build_procnode_from_lhi(
    id: int,
    name: str,
    topology: LhiTopology,
    latencies: LhiLatencies,
    network_info: NetworkInfo,
    network_lhi: LhiNetworkInfo,
) -> ProcNode:
    qprocessor = build_qprocessor_from_topology(f"{name}_processor", topology)
    network_ehi = LhiConverter.network_to_ehi(network_lhi)
    return ProcNode(
        name=name,
        node_id=id,
        network_info=network_info,
        qprocessor=qprocessor,
        qdevice_topology=topology,
        latencies=latencies,
        ntf_interface=GenericToVanillaInterface(),
        network_ehi=network_ehi,
    )


def build_network_from_lhi(
    procnode_infos: List[LhiProcNodeInfo],
    network_info: NetworkInfo,
    network_lhi: LhiNetworkInfo,
) -> ProcNodeNetwork:
    procnodes: Dict[str, ProcNode] = {}

    for info in procnode_infos:
        procnode = build_procnode_from_lhi(
            info.id, info.name, info.topology, info.latencies, network_info, network_lhi
        )
        procnodes[info.name] = procnode

    ns_nodes = [procnode.node for procnode in procnodes.values()]
    entdistcomp = EntDistComponent(network_info)
    entdist = EntDist(nodes=ns_nodes, network_info=network_info, comp=entdistcomp)

    for ([n1, n2], link_info) in network_lhi.links.items():
        entdist.add_sampler(n1, n2, link_info)

    for (_, s1), (_, s2) in itertools.combinations(procnodes.items(), 2):
        s1.connect_to(s2)

    for name, procnode in procnodes.items():
        procnode.node.entdist_out_port.connect(entdistcomp.node_in_port(name))
        procnode.node.entdist_in_port.connect(entdistcomp.node_out_port(name))

    return ProcNodeNetwork(procnodes, entdist)



from __future__ import annotations

from dataclasses import dataclass
from typing import Generator

from pydynaa import EventExpression
from qoala.runtime.environment import LocalEnvironment
from qoala.runtime.message import Message
from qoala.sim.componentprot import ComponentProtocol, PortListener
from qoala.sim.events import (
    EVENT_WAIT,
    SIGNAL_ENTD_NSTK_MSG,
    SIGNAL_HOST_NSTK_MSG,
    SIGNAL_MEMORY_FREED,
    SIGNAL_NSTK_NSTK_MSG,
    SIGNAL_QNOS_NSTK_MSG,
)
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack.netstackcomp import NetstackComponent
from qoala.sim.qdevice import QDevice


@dataclass
class NetstackLatencies:
    netstack_peer_latency: float = 0  # processing time for messages from remote node

    @classmethod
    def all_zero(cls) -> NetstackLatencies:
        # NOTE: can also just use NetstackLatencies() which will default all values to 0
        # However, using this classmethod makes this behavior more explicit and clear.
        return NetstackLatencies(0)


class NetstackInterface(ComponentProtocol):
    """NetSquid protocol representing the QNodeOS network stack."""

    def __init__(
        self,
        comp: NetstackComponent,
        local_env: LocalEnvironment,
        qdevice: QDevice,
        memmgr: MemoryManager,
    ) -> None:
        """Network stack protocol constructor. Typically created indirectly through
        constructing a `Qnos` instance.

        :param comp: NetSquid component representing the network stack
        :param qnos: `Qnos` protocol that owns this protocol
        """
        super().__init__(name=f"{comp.name}_protocol", comp=comp)
        self._comp = comp
        self._qdevice = qdevice
        self._local_env = local_env
        self._memmgr = memmgr

        self.add_listener(
            "host",
            PortListener(self._comp.host_in_port, SIGNAL_HOST_NSTK_MSG),
        )

        self.add_listener(
            "qnos",
            PortListener(self._comp.qnos_in_port, SIGNAL_QNOS_NSTK_MSG),
        )

        self.add_listener(
            "qnos_mem",
            PortListener(self._comp.qnos_mem_in_port, SIGNAL_MEMORY_FREED),
        )

        self.add_listener(
            "entdist",
            PortListener(self._comp.entdist_in_port, SIGNAL_ENTD_NSTK_MSG),
        )

        for peer in self._local_env.get_all_other_node_names():
            self.add_listener(
                f"peer_{peer}",
                PortListener(
                    self._comp.peer_in_port(peer), f"{SIGNAL_NSTK_NSTK_MSG}_{peer}"
                ),
            )

    def send_host_msg(self, msg: Message) -> None:
        """Send a message to the host."""
        self._comp.host_out_port.tx_output(msg)

    def receive_host_msg(self) -> Generator[EventExpression, None, Message]:
        """Receive a message from the host. Block until there is at least one
        message."""
        return (yield from self._receive_msg("host", SIGNAL_HOST_NSTK_MSG))

    def send_qnos_msg(self, msg: Message) -> None:
        """Send a message to the processor."""
        self._comp.qnos_out_port.tx_output(msg)

    def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
        """Receive a message from the processor. Block until there is at least one
        message."""
        return (yield from self._receive_msg("qnos", SIGNAL_QNOS_NSTK_MSG))

    def send_entdist_msg(self, msg: Message) -> None:
        """Send a message to the Entdist."""
        self._comp.entdist_out_port.tx_output(msg)

    def receive_entdist_msg(self) -> Generator[EventExpression, None, Message]:
        """Receive a message from the Entdist. Block until there is at least one
        message."""
        return (yield from self._receive_msg("entdist", SIGNAL_ENTD_NSTK_MSG))

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        """Send a message to the network stack of the other node.

        NOTE: for now we assume there is only one other node, which is 'the' peer."""
        self._comp.peer_out_port(peer).tx_output(msg)

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        """Receive a message from the network stack of the other node. Block until
        there is at least one message.

        NOTE: for now we assume there is only one other node, which is 'the' peer."""
        return (
            yield from self._receive_msg(
                f"peer_{peer}", f"{SIGNAL_NSTK_NSTK_MSG}_{peer}"
            )
        )

    def await_memory_freed_signal(
        self, pid: int, virt_id: int
    ) -> Generator[EventExpression, None, None]:
        # TODO: use pid and virt_id?
        yield from self._receive_msg("qnos_mem", SIGNAL_MEMORY_FREED)

    @property
    def qdevice(self) -> QDevice:
        return self._qdevice

    @property
    def node_id(self) -> int:
        return self._qdevice.node.ID  # type: ignore

    @property
    def memmgr(self) -> MemoryManager:
        return self._memmgr

    def remote_id_to_peer_name(self, remote_id: int) -> str:
        return self._local_env.get_network_info().get_nodes()[remote_id]

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr



from typing import Any, Callable, List, Optional

class Entity:
    pass

class EventType:
    def __init__(self, name: str, description: str) -> None: ...

class EventHandler:
    def __init__(
        self,
        callback_function: Callable,
        identifier: str = ...,
        safe_guards: List[Entity] = ...,
        priority: int = ...,
    ) -> None: ...

class EventExpression:
    def __init__(
        self,
        source: Optional[Entity] = ...,
        event_type: Optional[EventType] = ...,
        event_id: int = ...,
    ) -> None: ...
    @property
    def atomic_type(self) -> Optional[Any]: ...

class Event:
    @property
    def id(self) -> int: ...
    @property
    def source(self) -> Entity: ...
    @property
    def type(self) -> EventType: ...
    def unschedule(self) -> None: ...

class Protocol:
    def send_signal(self, signal: str) -> None: ...

class QuantumProcessor:
    @property
    def num_positions(self) -> int: ...



import os
from typing import Optional

import netsquid as ns
from netqasm.lang.instr.flavour import core

from qoala.lang.ehi import EhiNodeInfo, UnitModule
from qoala.lang.hostlang import BasicBlockType
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.environment import NetworkInfo
from qoala.runtime.lhi import (
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiProcNodeInfo,
    LhiTopologyBuilder,
)
from qoala.runtime.program import ProgramInput, ProgramInstance
from qoala.runtime.schedule import (
    LinkSlotInfo,
    QcSlotInfo,
    TaskSchedule,
    TaskScheduleEntry,
)
from qoala.runtime.task import BlockTask, TaskCreator, TaskExecutionMode
from qoala.sim.build import build_network_from_lhi
from qoala.sim.network import ProcNodeNetwork


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


def setup_network() -> ProcNodeNetwork:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits=3)
    latencies = LhiLatencies(
        host_instr_time=1000, qnos_instr_time=2000, host_peer_latency=3000
    )
    link_info = LhiLinkInfo.perfect(duration=20_000)

    alice_lhi = LhiProcNodeInfo(
        name="alice", id=0, topology=topology, latencies=latencies
    )
    network_lhi = LhiNetworkInfo.fully_connected([0, 1], link_info)
    network_info = NetworkInfo.with_nodes({0: "alice", 1: "bob"})
    bob_lhi = LhiProcNodeInfo(name="bob", id=1, topology=topology, latencies=latencies)
    return build_network_from_lhi([alice_lhi, bob_lhi], network_info, network_lhi)


def instantiate(
    program: QoalaProgram,
    ehi: EhiNodeInfo,
    pid: int = 0,
    inputs: Optional[ProgramInput] = None,
) -> ProgramInstance:
    unit_module = UnitModule.from_full_ehi(ehi)

    if inputs is None:
        inputs = ProgramInput.empty()

    return ProgramInstance(
        pid,
        program,
        inputs,
        unit_module=unit_module,
        block_tasks=[],
    )


CL = BasicBlockType.CL
CC = BasicBlockType.CC
QL = BasicBlockType.QL
QC = BasicBlockType.QC


def test_consecutive():
    pid = 0
    tasks = [
        BlockTask(pid, "blk_host0", CL, 1000),
        BlockTask(pid, "blk_recv", CC, 1000),
        BlockTask(pid, "blk_add_one", QL, 1000),
        BlockTask(pid, "blk_epr_md_1", QC, 1000),
        BlockTask(pid, "blk_host1", CL, 1000),
    ]

    schedule = TaskSchedule.consecutive(tasks)

    assert schedule.entries == [
        TaskScheduleEntry(tasks[0], None, prev=None),
        TaskScheduleEntry(tasks[1], None, prev=None),
        TaskScheduleEntry(tasks[2], None, prev=tasks[1]),  # CPU -> QPU
        TaskScheduleEntry(tasks[3], None, prev=None),
        TaskScheduleEntry(tasks[4], None, prev=tasks[3]),  # QPU -> CPU
    ]


def test_consecutive_qc_slots():
    pid = 0
    tasks = [
        BlockTask(pid, "blk_host0", CL, 1000),
        BlockTask(pid, "blk_recv", CC, 10000),
        BlockTask(pid, "blk_add_one", QL, 1000),
        BlockTask(pid, "blk_epr_md_1", QC, 1000, remote_id=0),
        BlockTask(pid, "blk_host1", CL, 1000),
    ]

    schedule = TaskSchedule.consecutive(
        tasks, qc_slot_info=QcSlotInfo({0: LinkSlotInfo(0, 100, 50_000)})
    )

    assert schedule.entries == [
        TaskScheduleEntry(tasks[0], timestamp=None, prev=None),
        TaskScheduleEntry(tasks[1], timestamp=None, prev=None),
        TaskScheduleEntry(tasks[2], timestamp=None, prev=tasks[1]),  # CPU -> QPU
        TaskScheduleEntry(tasks[3], timestamp=50_000, prev=None),  # QC task
        TaskScheduleEntry(tasks[4], timestamp=None, prev=tasks[3]),  # QPU -> CPU
    ]


def test_consecutive_timestamps():
    pid = 0

    tasks = [
        BlockTask(pid, "blk_host0", CL, 1000),
        BlockTask(pid, "blk_lr0", QL, 5000),
        BlockTask(pid, "blk_host1", CL, 200),
        BlockTask(pid, "blk_rr0", QC, 30_000, remote_id=0),
        BlockTask(pid, "blk_host2", CL, 4000),
    ]

    schedule = TaskSchedule.consecutive_timestamps(tasks)

    assert schedule.entries == [
        TaskScheduleEntry(tasks[0], 0),
        TaskScheduleEntry(tasks[1], 1000),
        TaskScheduleEntry(tasks[2], 1000 + 5000),
        TaskScheduleEntry(tasks[3], 1000 + 5000 + 200),
        TaskScheduleEntry(tasks[4], 1000 + 5000 + 200 + 30_000),
    ]

    assert schedule.cpu_schedule.entries == [
        TaskScheduleEntry(tasks[0], 0),
        TaskScheduleEntry(tasks[2], 1000 + 5000),
        TaskScheduleEntry(tasks[4], 1000 + 5000 + 200 + 30_000),
    ]
    assert schedule.qpu_schedule.entries == [
        TaskScheduleEntry(tasks[1], 1000),
        TaskScheduleEntry(tasks[3], 1000 + 5000 + 200),
    ]


def test_host_program():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program = load_program("test_scheduling_alice.iqoala")
    pid = 0
    instance = instantiate(program, alice.local_ehi, pid)

    alice.scheduler.submit_program_instance(instance)
    bob.scheduler.submit_program_instance(instance)

    cpu_schedule = TaskSchedule.consecutive(
        [
            BlockTask(pid, "blk_host0", CL),
            BlockTask(pid, "blk_host1", CL),
        ]
    )
    alice.scheduler.upload_cpu_schedule(cpu_schedule)
    bob.scheduler.upload_cpu_schedule(cpu_schedule)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    assert ns.sim_time() == 3 * alice.local_ehi.latencies.host_instr_time
    alice.memmgr.get_process(pid).host_mem.read("var_z") == 9
    bob.memmgr.get_process(pid).host_mem.read("var_z") == 9


def test_lr_program():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program = load_program("test_scheduling_alice.iqoala")
    pid = 0
    instance = instantiate(program, alice.local_ehi, pid)

    alice.scheduler.submit_program_instance(instance)
    bob.scheduler.submit_program_instance(instance)

    host_instr_time = alice.local_ehi.latencies.host_instr_time
    schedule = TaskSchedule.consecutive(
        [
            BlockTask(pid, "blk_host2", CL),
            BlockTask(pid, "blk_add_one", QL),
        ]
    )
    alice.scheduler.upload_schedule(schedule)
    bob.scheduler.upload_schedule(schedule)
    print(schedule)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    host_instr_time = alice.local_ehi.latencies.host_instr_time
    qnos_instr_time = alice.local_ehi.latencies.qnos_instr_time
    expected_duration = host_instr_time + 5 * qnos_instr_time
    assert ns.sim_time() == expected_duration
    alice.memmgr.get_process(pid).host_mem.read("y") == 4
    bob.memmgr.get_process(pid).host_mem.read("y") == 4


def test_epr_md_1():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    alice.scheduler.submit_program_instance(instance_alice)
    bob.scheduler.submit_program_instance(instance_bob)

    qpu_schedule = TaskSchedule([TaskScheduleEntry(BlockTask(pid, "blk_epr_md_1", QC))])
    alice.scheduler.upload_qpu_schedule(qpu_schedule)
    bob.scheduler.upload_qpu_schedule(qpu_schedule)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    expected_duration = alice.network_ehi.get_link(0, 1).duration
    assert ns.sim_time() == expected_duration
    alice_outcome = alice.memmgr.get_process(pid).host_mem.read("m")
    bob_outcome = bob.memmgr.get_process(pid).host_mem.read("m")
    assert alice_outcome == bob_outcome


def test_epr_md_2():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    alice.scheduler.submit_program_instance(instance_alice)
    bob.scheduler.submit_program_instance(instance_bob)

    qpu_schedule = TaskSchedule([TaskScheduleEntry(BlockTask(pid, "blk_epr_md_2", QC))])
    alice.scheduler.upload_qpu_schedule(qpu_schedule)
    bob.scheduler.upload_qpu_schedule(qpu_schedule)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    expected_duration = alice.network_ehi.get_link(0, 1).duration * 2
    assert ns.sim_time() == expected_duration
    alice_mem = alice.memmgr.get_process(pid).host_mem
    bob_mem = bob.memmgr.get_process(pid).host_mem
    alice_outcomes = [alice_mem.read("m0"), alice_mem.read("m1")]
    bob_outcomes = [bob_mem.read("m0"), bob_mem.read("m1")]
    assert alice_outcomes == bob_outcomes


def test_epr_ck_1():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    alice.scheduler.submit_program_instance(instance_alice)
    bob.scheduler.submit_program_instance(instance_bob)

    qpu_schedule = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(pid, "blk_epr_ck_1", QC)),
            TaskScheduleEntry(BlockTask(pid, "blk_meas_q0", QL)),
        ]
    )
    alice.scheduler.upload_qpu_schedule(qpu_schedule)
    bob.scheduler.upload_qpu_schedule(qpu_schedule)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    # subrt meas_q0 has 3 classical instructions + 1 meas instruction
    subrt_class_time = 3 * alice.local_ehi.latencies.qnos_instr_time
    subrt_meas_time = alice.local_ehi.find_single_gate(0, core.MeasInstruction).duration
    subrt_time = subrt_class_time + subrt_meas_time
    expected_duration = alice.network_ehi.get_link(0, 1).duration + subrt_time
    assert ns.sim_time() == expected_duration
    alice_outcome = alice.memmgr.get_process(pid).host_mem.read("p")
    bob_outcome = bob.memmgr.get_process(pid).host_mem.read("p")
    assert alice_outcome == bob_outcome


def test_epr_ck_2():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    alice.scheduler.submit_program_instance(instance_alice)
    bob.scheduler.submit_program_instance(instance_bob)

    qpu_schedule = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(pid, "blk_epr_ck_2", QC)),
            TaskScheduleEntry(BlockTask(pid, "blk_meas_q0_q1", QL)),
        ]
    )
    alice.scheduler.upload_qpu_schedule(qpu_schedule)
    bob.scheduler.upload_qpu_schedule(qpu_schedule)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    # subrt meas_q0_q1 has 6 classical instructions + 2 meas instruction
    subrt_class_time = 6 * alice.local_ehi.latencies.qnos_instr_time
    meas_time = alice.local_ehi.find_single_gate(0, core.MeasInstruction).duration
    subrt_time = subrt_class_time + 2 * meas_time
    epr_time = alice.network_ehi.get_link(0, 1).duration
    expected_duration = 2 * epr_time + subrt_time
    assert ns.sim_time() == expected_duration
    alice_mem = alice.memmgr.get_process(pid).host_mem
    bob_mem = bob.memmgr.get_process(pid).host_mem
    alice_outcomes = [alice_mem.read("p0"), alice_mem.read("p1")]
    bob_outcomes = [bob_mem.read("p0"), bob_mem.read("p1")]
    assert alice_outcomes == bob_outcomes


def test_cc():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    alice.scheduler.submit_program_instance(instance_alice)
    bob.scheduler.submit_program_instance(instance_bob)

    assert alice.local_ehi.latencies.host_peer_latency == 3000
    assert alice.local_ehi.latencies.host_instr_time == 1000

    schedule_alice = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(pid, "blk_prep_cc", CL), 0),
            TaskScheduleEntry(BlockTask(pid, "blk_send", CL), 2000),
            TaskScheduleEntry(BlockTask(pid, "blk_host1", CL), 10000),
        ]
    )
    schedule_bob = TaskSchedule(
        [
            TaskScheduleEntry(BlockTask(pid, "blk_prep_cc", CL), 0),
            TaskScheduleEntry(BlockTask(pid, "blk_recv", CC), 3000),
            TaskScheduleEntry(BlockTask(pid, "blk_host1", CL), 10000),
        ]
    )
    alice.scheduler.upload_cpu_schedule(schedule_alice)
    bob.scheduler.upload_cpu_schedule(schedule_bob)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    # assert ns.sim_time() == expected_duration
    alice_mem = alice.memmgr.get_process(pid).host_mem
    bob_mem = bob.memmgr.get_process(pid).host_mem
    assert bob_mem.read("msg") == 25
    assert alice_mem.read("var_z") == 9
    assert bob_mem.read("var_z") == 9


def test_full_program():
    network = setup_network()
    alice = network.nodes["alice"]
    bob = network.nodes["bob"]

    program_alice = load_program("test_scheduling_alice.iqoala")
    program_bob = load_program("test_scheduling_bob.iqoala")
    pid = 0
    inputs_alice = ProgramInput({"bob_id": 1})
    inputs_bob = ProgramInput({"alice_id": 0})
    instance_alice = instantiate(program_alice, alice.local_ehi, pid, inputs_alice)
    instance_bob = instantiate(program_bob, bob.local_ehi, pid, inputs_bob)

    alice.scheduler.submit_program_instance(instance_alice)
    bob.scheduler.submit_program_instance(instance_bob)

    task_creator = TaskCreator(mode=TaskExecutionMode.ROUTINE_ATOMIC)
    tasks_alice = task_creator.from_program(
        program_alice, pid, alice.local_ehi, alice.network_ehi
    )
    tasks_bob = task_creator.from_program(
        program_bob, pid, bob.local_ehi, bob.network_ehi
    )
    schedule_alice = TaskSchedule.consecutive(tasks_alice)
    schedule_bob = TaskSchedule.consecutive(tasks_bob)

    alice.scheduler.upload_schedule(schedule_alice)
    bob.scheduler.upload_schedule(schedule_bob)

    ns.sim_reset()
    network.start()
    ns.sim_run()

    alice_mem = alice.memmgr.get_process(pid).host_mem
    bob_mem = bob.memmgr.get_process(pid).host_mem
    alice_outcomes = [alice_mem.read("p0"), alice_mem.read("p1")]
    bob_outcomes = [bob_mem.read("p0"), bob_mem.read("p1")]
    assert alice_outcomes == bob_outcomes


if __name__ == "__main__":
    test_consecutive()
    test_consecutive_qc_slots()
    test_consecutive_timestamps()
    test_host_program()
    test_lr_program()
    test_epr_md_1()
    test_epr_md_2()
    test_epr_ck_1()
    test_epr_ck_2()
    test_cc()
    test_full_program()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List

from qoala.runtime.sharedmem import MemAddr


@dataclass
class Message:
    content: Any


@dataclass
class LrCallTuple:
    routine_name: str
    input_addr: MemAddr
    result_addr: MemAddr


@dataclass
class RrCallTuple:
    routine_name: str
    input_addr: MemAddr
    result_addr: MemAddr
    cb_input_addrs: List[MemAddr]
    cb_output_addrs: List[MemAddr]

    @classmethod
    def no_alloc(cls, name: str) -> RrCallTuple:
        return RrCallTuple(name, MemAddr(0), MemAddr(0), [], [])



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Optional

from netqasm.lang.instr import core

from qoala.lang.ehi import EhiNetworkInfo, EhiNodeInfo
from qoala.lang.hostlang import (
    BasicBlockType,
    ReceiveCMsgOp,
    RunRequestOp,
    RunSubroutineOp,
)
from qoala.lang.program import QoalaProgram
from qoala.lang.routine import LocalRoutine


class TaskExecutionMode(Enum):
    ROUTINE_ATOMIC = 0
    ROUTINE_SPLIT = auto()


@dataclass(eq=True, frozen=True)
class BlockTask:
    pid: int
    block_name: str
    typ: BasicBlockType
    duration: Optional[float] = None
    max_time: Optional[float] = None
    remote_id: Optional[int] = None

    def __str__(self) -> str:
        return f"{self.block_name} ({self.typ.name}), dur={self.duration}"


class TaskCreator:
    def __init__(self, mode: TaskExecutionMode) -> None:
        self._mode = mode

    def from_program(
        self,
        program: QoalaProgram,
        pid: int,
        ehi: Optional[EhiNodeInfo] = None,
        network_ehi: Optional[EhiNetworkInfo] = None,
        remote_id: Optional[int] = None,
    ) -> List[BlockTask]:
        if self._mode == TaskExecutionMode.ROUTINE_ATOMIC:
            return self._from_program_routine_atomic(
                program, pid, ehi, network_ehi, remote_id
            )
        else:
            raise NotImplementedError

    def _from_program_routine_atomic(
        self,
        program: QoalaProgram,
        pid: int,
        ehi: Optional[EhiNodeInfo] = None,
        network_ehi: Optional[EhiNetworkInfo] = None,
        remote_id: Optional[int] = None,
    ) -> List[BlockTask]:
        tasks: List[BlockTask] = []

        for block in program.blocks:
            if block.typ == BasicBlockType.CL:
                if ehi is not None:
                    duration = ehi.latencies.host_instr_time * len(block.instructions)
                else:
                    duration = None
                cputask = BlockTask(pid, block.name, block.typ, duration)
                tasks.append(cputask)
            elif block.typ == BasicBlockType.CC:
                assert len(block.instructions) == 1
                instr = block.instructions[0]
                assert isinstance(instr, ReceiveCMsgOp)
                if ehi is not None:
                    duration = ehi.latencies.host_peer_latency
                else:
                    duration = None
                cputask = BlockTask(pid, block.name, block.typ, duration)
                tasks.append(cputask)
            elif block.typ == BasicBlockType.QL:
                assert len(block.instructions) == 1
                instr = block.instructions[0]
                assert isinstance(instr, RunSubroutineOp)
                if ehi is not None:
                    local_routine = program.local_routines[instr.subroutine]
                    duration = self._compute_lr_duration(ehi, local_routine)
                else:
                    duration = None
                qputask = BlockTask(pid, block.name, block.typ, duration)
                tasks.append(qputask)
            elif block.typ == BasicBlockType.QC:
                assert len(block.instructions) == 1
                instr = block.instructions[0]
                assert isinstance(instr, RunRequestOp)
                if network_ehi is not None:
                    # TODO: refactor!!
                    epr_time = list(network_ehi.links.values())[0].duration
                    req_routine = program.request_routines[instr.req_routine]
                    duration = epr_time * req_routine.request.num_pairs
                else:
                    duration = None

                qputask = BlockTask(
                    pid, block.name, block.typ, duration, remote_id=remote_id
                )
                tasks.append(qputask)

        return tasks

    def _compute_lr_duration(self, ehi: EhiNodeInfo, routine: LocalRoutine) -> float:
        duration = 0.0
        # TODO: refactor this
        for instr in routine.subroutine.instructions:
            if (
                type(instr)
                in [
                    core.SetInstruction,
                    core.StoreInstruction,
                    core.LoadInstruction,
                    core.LeaInstruction,
                ]
                or isinstance(instr, core.BranchBinaryInstruction)
                or isinstance(instr, core.BranchUnaryInstruction)
                or isinstance(instr, core.JmpInstruction)
                or isinstance(instr, core.ClassicalOpInstruction)
                or isinstance(instr, core.ClassicalOpModInstruction)
            ):
                duration += ehi.latencies.qnos_instr_time
            else:
                max_duration = -1.0
                # TODO: gate duration depends on which qubit!!
                # currently we always take the worst case scenario but this is not ideal
                for i in ehi.single_gate_infos.keys():
                    if info := ehi.find_single_gate(i, type(instr)):
                        max_duration = max(max_duration, info.duration)

                for multi in ehi.multi_gate_infos.keys():
                    if info := ehi.find_multi_gate(multi.qubit_ids, type(instr)):
                        max_duration = max(max_duration, info.duration)

                if max_duration != -1:
                    duration += max_duration
                else:
                    raise RuntimeError
        return duration



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Optional, Union

from netqasm.lang.operand import Template

IqoalaValue = Union[int, Template, str]


class IqoalaInstructionType(Enum):
    CC = 0
    CL = auto()
    QC = auto()
    QL = auto()


class IqoalaAttribute:
    def __init__(self, value: IqoalaValue) -> None:
        self._value = value

    @property
    def value(self) -> IqoalaValue:
        return self._value


class IqoalaVector:
    def __init__(self, values: List[str]) -> None:
        self._values = values

    @property
    def values(self) -> List[str]:
        return self._values

    def __str__(self) -> str:
        return f"vec<{','.join(v for v in self.values)}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, IqoalaVector):
            return NotImplemented
        return self.values == other.values


class ClassicalIqoalaOp:
    OP_NAME: str = None  # type: ignore
    TYP: IqoalaInstructionType = None  # type: ignore

    def __init__(
        self,
        arguments: Optional[Union[List[str], List[IqoalaVector]]] = None,
        results: Optional[Union[List[str], IqoalaVector]] = None,
        attributes: Optional[List[IqoalaValue]] = None,
    ) -> None:
        # TODO: support list of strs and vectors
        # currently not needed and confuses mypy
        self._arguments: Union[List[str], List[IqoalaVector]]
        self._results: Union[List[str], IqoalaVector]
        self._attributes: List[IqoalaValue]

        if arguments is None:
            self._arguments = []  # type: ignore
        else:
            self._arguments = arguments

        if results is None:
            self._results = []
        elif isinstance(results, list):
            # List of ints
            self._results = results
        else:
            assert isinstance(results, IqoalaVector)
            self._results = results

        if attributes is None:
            self._attributes = []
        else:
            self._attributes = attributes

    def __str__(self) -> str:
        if isinstance(self.results, list):
            results = ", ".join(str(r) for r in self.results)
        else:
            assert isinstance(self.results, IqoalaVector)
            results = str(self.results)
        args = ", ".join(str(a) for a in self.arguments)
        attrs = ", ".join(str(a) for a in self.attributes)
        s = ""
        if len(results) > 0:
            s += f"{results} = "

        s += f"{self.op_name}({args})"

        if len(attrs) > 0:
            s += f" : {attrs}"
        return s

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ClassicalIqoalaOp):
            return NotImplemented
        return (
            self.results == other.results
            and self.arguments == other.arguments
            and self.attributes == other.attributes
        )

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ) -> ClassicalIqoalaOp:
        raise NotImplementedError

    @property
    def op_name(self) -> str:
        return self.__class__.OP_NAME  # type: ignore

    @property
    def arguments(self) -> Union[List[str], List[IqoalaVector]]:
        return self._arguments

    @property
    def results(self) -> Union[List[str], IqoalaVector]:
        return self._results

    @property
    def attributes(self) -> List[IqoalaValue]:
        return self._attributes


class AssignCValueOp(ClassicalIqoalaOp):
    OP_NAME = "assign_cval"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: str, value: IqoalaValue) -> None:
        super().__init__(results=[result], attributes=[value])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 0
        assert attr is not None
        return cls(result, attr)


class SendCMsgOp(ClassicalIqoalaOp):
    OP_NAME = "send_cmsg"
    TYP = IqoalaInstructionType.CC

    def __init__(self, csocket: str, value: str) -> None:
        # args:
        #   csocket (int): ID of csocket
        #   value (str): name of variable holding the value to send
        super().__init__(arguments=[csocket, value])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is None
        assert len(args) == 2
        assert attr is None
        return cls(args[0], args[1])


class ReceiveCMsgOp(ClassicalIqoalaOp):
    OP_NAME = "recv_cmsg"
    TYP = IqoalaInstructionType.CC

    def __init__(self, csocket: str, result: str) -> None:
        super().__init__(arguments=[csocket], results=[result])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 1
        assert attr is None
        return cls(args[0], result)


class AddCValueOp(ClassicalIqoalaOp):
    OP_NAME = "add_cval_c"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: str, value0: str, value1: str) -> None:
        super().__init__(arguments=[value0, value1], results=[result])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 2
        assert attr is None
        return cls(result, args[0], args[1])


class MultiplyConstantCValueOp(ClassicalIqoalaOp):
    OP_NAME = "mult_const"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: str, value0: str, const: IqoalaValue) -> None:
        # result = value0 * const
        super().__init__(arguments=[value0], attributes=[const], results=[result])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 1
        assert attr is not None
        return cls(result, args[0], attr)


class BitConditionalMultiplyConstantCValueOp(ClassicalIqoalaOp):
    OP_NAME = "bcond_mult_const"
    TYP = IqoalaInstructionType.CL

    def __init__(self, result: str, value0: str, cond: str, const: IqoalaValue) -> None:
        # if const == 1:
        #   result = value0 * const
        # else:
        #   result = value0
        super().__init__(arguments=[value0, cond], attributes=[const], results=[result])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is not None
        assert len(args) == 2
        assert attr is not None
        return cls(result, args[0], args[1], attr)


class RunSubroutineOp(ClassicalIqoalaOp):
    OP_NAME = "run_subroutine"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self, result: Optional[IqoalaVector], values: IqoalaVector, subrt: str
    ) -> None:
        super().__init__(results=result, arguments=[values], attributes=[subrt])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        if result is not None:
            assert isinstance(result, IqoalaVector)
        assert len(args) == 1
        assert isinstance(args[0], IqoalaVector)
        assert isinstance(attr, str)
        return cls(result, args[0], attr)

    @property
    def subroutine(self) -> str:
        assert isinstance(self.attributes[0], str)
        return self.attributes[0]

    def __str__(self) -> str:
        return super().__str__()


class RunRequestOp(ClassicalIqoalaOp):
    OP_NAME = "run_request"
    TYP = IqoalaInstructionType.CL

    def __init__(
        self, result: Optional[IqoalaVector], values: IqoalaVector, routine: str
    ) -> None:
        super().__init__(results=result, arguments=[values], attributes=[routine])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        if result is not None:
            assert isinstance(result, IqoalaVector)
        assert len(args) == 1
        assert isinstance(args[0], IqoalaVector)
        assert isinstance(attr, str)
        return cls(result, args[0], attr)

    @property
    def req_routine(self) -> str:
        assert isinstance(self.attributes[0], str)
        return self.attributes[0]

    def __str__(self) -> str:
        return super().__str__()


class ReturnResultOp(ClassicalIqoalaOp):
    OP_NAME = "return_result"
    TYP = IqoalaInstructionType.CL

    def __init__(self, value: str) -> None:
        super().__init__(arguments=[value])

    @classmethod
    def from_generic_args(
        cls, result: Optional[str], args: List[str], attr: Optional[IqoalaValue]
    ):
        assert result is None
        assert len(args) == 1
        assert attr is None
        return cls(args[0])


class BasicBlockType(Enum):
    CL = 0
    CC = auto()
    QL = auto()
    QC = auto()


@dataclass
class BasicBlockAnnotations:
    deadline: int


@dataclass
class BasicBlock:
    name: str
    typ: BasicBlockType
    instructions: List[ClassicalIqoalaOp]

    def __str__(self) -> str:
        s = f"^{self.name} {{type = {self.typ.name}}}:\n"
        return s + "\n".join("    " + str(i) for i in self.instructions)



from __future__ import annotations

import logging
from typing import Dict, Generator, List, Optional

from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo, EhiNodeInfo
from qoala.runtime.environment import LocalEnvironment
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.program import (
    BatchInfo,
    BatchResult,
    ProgramBatch,
    ProgramInstance,
    ProgramResult,
)
from qoala.runtime.schedule import TaskSchedule
from qoala.runtime.task import BlockTask, TaskCreator, TaskExecutionMode
from qoala.sim.driver import CpuDriver, QpuDriver
from qoala.sim.eprsocket import EprSocket
from qoala.sim.events import EVENT_WAIT
from qoala.sim.host.csocket import ClassicalSocket
from qoala.sim.host.host import Host
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack import Netstack
from qoala.sim.process import QoalaProcess
from qoala.sim.qnos import Qnos
from qoala.util.logging import LogManager


class Scheduler(Protocol):
    def __init__(
        self,
        node_name: str,
        host: Host,
        qnos: Qnos,
        netstack: Netstack,
        memmgr: MemoryManager,
        local_env: LocalEnvironment,
        local_ehi: EhiNodeInfo,
        network_ehi: EhiNetworkInfo,
    ) -> None:
        super().__init__(name=f"{node_name}_scheduler")

        self._node_name = node_name

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({node_name})"
        )

        self._host = host
        self._qnos = qnos
        self._netstack = netstack
        self._memmgr = memmgr
        self._local_env = local_env
        self._local_ehi = local_ehi
        self._network_ehi = network_ehi

        self._prog_instance_counter: int = 0
        self._batch_counter: int = 0
        self._batches: Dict[int, ProgramBatch] = {}  # batch ID -> batch
        self._prog_results: Dict[int, ProgramResult] = {}  # program ID -> result
        self._batch_results: Dict[int, BatchResult] = {}  # batch ID -> result

        self._block_schedule: Optional[TaskSchedule] = None

        self._cpudriver = CpuDriver(node_name, host.processor, memmgr)
        self._qpudriver = QpuDriver(
            node_name, host.processor, qnos.processor, netstack.processor, memmgr
        )

        self._cpudriver.set_other_driver(self._qpudriver)
        self._qpudriver.set_other_driver(self._cpudriver)

    @property
    def host(self) -> Host:
        return self._host

    @property
    def qnos(self) -> Qnos:
        return self._qnos

    @property
    def netstack(self) -> Netstack:
        return self._netstack

    @property
    def memmgr(self) -> MemoryManager:
        return self._memmgr

    @property
    def cpudriver(self) -> CpuDriver:
        return self._cpudriver

    @property
    def qpudriver(self) -> QpuDriver:
        return self._qpudriver

    @property
    def block_schedule(self) -> TaskSchedule:
        assert self._block_schedule is not None
        return self._block_schedule

    def submit_batch(self, batch_info: BatchInfo) -> None:
        prog_instances: List[ProgramInstance] = []

        network_info = self._local_env.get_network_info()

        for i in range(batch_info.num_iterations):
            pid = self._prog_instance_counter
            task_creator = TaskCreator(mode=TaskExecutionMode.ROUTINE_ATOMIC)
            # TODO: allow multiple remote nodes in single program??
            remote_names = list(batch_info.program.meta.csockets.values())
            if len(remote_names) > 0:
                remote_name = list(batch_info.program.meta.csockets.values())[0]
                remote_id = network_info.get_node_id(remote_name)
            else:
                remote_id = None
            block_tasks = task_creator.from_program(
                batch_info.program, pid, self._local_ehi, self._network_ehi, remote_id
            )

            instance = ProgramInstance(
                pid=pid,
                program=batch_info.program,
                inputs=batch_info.inputs[i],
                unit_module=batch_info.unit_module,
                block_tasks=block_tasks,
            )
            self._prog_instance_counter += 1
            prog_instances.append(instance)

        batch = ProgramBatch(
            batch_id=self._batch_counter, info=batch_info, instances=prog_instances
        )
        self._batches[batch.batch_id] = batch
        self._batch_counter += 1

    def get_batches(self) -> Dict[int, ProgramBatch]:
        return self._batches

    def create_process(self, prog_instance: ProgramInstance) -> QoalaProcess:
        prog_memory = ProgramMemory(prog_instance.pid)
        meta = prog_instance.program.meta

        csockets: Dict[int, ClassicalSocket] = {}
        for i, remote_name in meta.csockets.items():
            # TODO: check for already existing epr sockets
            csockets[i] = self.host.create_csocket(remote_name)

        epr_sockets: Dict[int, EprSocket] = {}
        for i, remote_name in meta.epr_sockets.items():
            network_info = self._local_env.get_network_info()
            remote_id = network_info.get_node_id(remote_name)
            # TODO: check for already existing epr sockets
            # TODO: fidelity
            epr_sockets[i] = EprSocket(i, remote_id, 1.0)

        result = ProgramResult(values={})

        return QoalaProcess(
            prog_instance=prog_instance,
            prog_memory=prog_memory,
            csockets=csockets,
            epr_sockets=epr_sockets,
            result=result,
        )

    def create_processes_for_batches(self) -> None:
        for batch in self._batches.values():
            for prog_instance in batch.instances:
                process = self.create_process(prog_instance)

                self.memmgr.add_process(process)
                self.initialize_process(process)

    def collect_batch_results(self) -> None:
        for batch_id, batch in self._batches.items():
            results: List[ProgramResult] = []
            for prog_instance in batch.instances:
                process = self.memmgr.get_process(prog_instance.pid)
                results.append(process.result)
            self._batch_results[batch_id] = BatchResult(batch_id, results)

    def get_batch_results(self) -> Dict[int, BatchResult]:
        self.collect_batch_results()
        return self._batch_results

    def initialize_process(self, process: QoalaProcess) -> None:
        # Write program inputs to host memory.
        self.host.processor.initialize(process)

        # TODO: rethink how and when Requests are instantiated
        # inputs = process.prog_instance.inputs
        # for req in process.get_all_requests().values():
        #     # TODO: support for other request parameters being templates?
        #     remote_id = req.request.remote_id
        #     if isinstance(remote_id, Template):
        #         req.request.remote_id = inputs.values[remote_id.name]

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr

    def upload_cpu_schedule(self, schedule: TaskSchedule) -> None:
        self._cpudriver.upload_schedule(schedule)

    def upload_qpu_schedule(self, schedule: TaskSchedule) -> None:
        self._qpudriver.upload_schedule(schedule)

    def upload_schedule(self, schedule: TaskSchedule) -> None:
        self._block_schedule = schedule
        self._cpudriver.upload_schedule(schedule.cpu_schedule)
        self._qpudriver.upload_schedule(schedule.qpu_schedule)

    def start(self) -> None:
        super().start()
        self._cpudriver.start()
        self._qpudriver.start()

    def stop(self) -> None:
        self._qpudriver.stop()
        self._cpudriver.stop()
        super().stop()

    def submit_program_instance(self, prog_instance: ProgramInstance) -> None:
        process = self.create_process(prog_instance)
        self.memmgr.add_process(process)
        self.initialize_process(process)

    def get_tasks_to_schedule(self) -> List[BlockTask]:
        all_tasks: List[BlockTask] = []

        for batch in self._batches.values():
            for inst in batch.instances:
                all_tasks.extend(inst.block_tasks)

        return all_tasks



from __future__ import annotations

from typing import Dict

from netsquid.components import QuantumProcessor
from netsquid.nodes.network import Network

from qoala.sim.entdist.entdist import EntDist
from qoala.sim.procnode import ProcNode


class ProcNodeNetwork(Network):
    """A network of `ProcNode`s connected by links, which are
    `MagicLinkLayerProtocol`s."""

    def __init__(self, nodes: Dict[str, ProcNode], entdist: EntDist) -> None:
        """ProcNodeNetwork constructor.

        :param nodes: dictionary of node name to `ProcNode` object representing
        that node
        :param links: list of link layer protocol objects. Each object internally
        contains the IDs of the two nodes that this link connects
        """
        self._nodes = nodes
        self._entdist = entdist

    @property
    def nodes(self) -> Dict[str, ProcNode]:
        return self._nodes

    @property
    def entdist(self) -> EntDist:
        return self._entdist

    @property
    def qdevices(self) -> Dict[str, QuantumProcessor]:
        return {name: node.qdevice for name, node in self._nodes.items()}

    def start_all_nodes(self) -> None:
        for node in self.nodes.values():
            node.start()

    def start_entdist(self) -> None:
        self.entdist.start()

    def start(self) -> None:
        self.start_entdist()
        self.start_all_nodes()



from __future__ import annotations

from typing import Dict, List, Optional, Tuple

from qoala.lang.ehi import EhiLinkInfo
from qoala.runtime.program import ProgramInstance


class NetworkInfo:
    """Static network info: node IDs. EPR links are managed by EhiNetworkInfo."""

    def __init__(self, nodes: Dict[int, str]) -> None:
        # node ID -> node name
        self._nodes = nodes

        self._global_schedule: Optional[List[int]] = None
        self._timeslot_len: Optional[int] = None

    @classmethod
    def with_nodes(cls, nodes: Dict[int, str]) -> NetworkInfo:
        return NetworkInfo(nodes)

    def get_nodes(self) -> Dict[int, str]:
        return self._nodes

    def get_node_id(self, name: str) -> int:
        for id, node_name in self._nodes.items():
            if node_name == name:
                return id
        raise ValueError

    def get_all_node_names(self) -> List[str]:
        return list(self._nodes.values())

    def set_nodes(self, nodes: Dict[int, str]) -> None:
        self._nodes = nodes

    def add_node(self, id: int, name: str) -> None:
        self._nodes[id] = name

    def get_links(self) -> Dict[Tuple[int, int], EhiLinkInfo]:
        return self._links

    def set_links(self, links: Dict[Tuple[int, int], EhiLinkInfo]) -> None:
        self._links = links

    def add_link(self, id1: int, id2: int, link: EhiLinkInfo) -> None:
        self._links[(id1, id2)] = link

    def set_global_schedule(self, schedule: List[int]) -> None:
        self._global_schedule = schedule

    def get_global_schedule(self) -> List[int]:
        assert self._global_schedule is not None
        return self._global_schedule

    def set_timeslot_len(self, len: int) -> None:
        self._timeslot_len = len

    def get_timeslot_len(self) -> int:
        assert self._timeslot_len is not None
        return self._timeslot_len


class LocalEnvironment:
    def __init__(
        self,
        network_info: NetworkInfo,
        node_id: int,
    ) -> None:
        self._network_info: NetworkInfo = network_info

        # node ID of self
        self._node_id: int = node_id

        self._programs: List[ProgramInstance] = []
        self._csockets: List[str] = []
        self._epr_sockets: List[str] = []

    def get_network_info(self) -> NetworkInfo:
        return self._network_info

    def get_node_id(self) -> int:
        return self._node_id

    def register_program(self, program: ProgramInstance) -> None:
        self._programs.append(program)

    def open_epr_socket(self) -> None:
        pass

    def get_all_node_names(self) -> List[str]:
        return self.get_network_info().get_all_node_names()

    def get_all_other_node_names(self) -> List[str]:
        return [
            name
            for id, name in self.get_network_info().get_nodes().items()
            if id != self._node_id
        ]



from dataclasses import dataclass
from typing import Dict

from qoala.lang.program import QoalaProgram
from qoala.lang.request import RequestRoutine
from qoala.lang.routine import LocalRoutine
from qoala.runtime.memory import HostMemory, ProgramMemory, QnosMemory
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.sharedmem import SharedMemory
from qoala.sim.eprsocket import EprSocket
from qoala.sim.host.csocket import ClassicalSocket


@dataclass
class QoalaProcess:
    prog_instance: ProgramInstance

    # Mutable
    prog_memory: ProgramMemory
    result: ProgramResult

    # Immutable
    csockets: Dict[int, ClassicalSocket]
    epr_sockets: Dict[int, EprSocket]

    def get_local_routine(self, name: str) -> LocalRoutine:
        return self.program.local_routines[name]

    def get_all_local_routines(self) -> Dict[str, LocalRoutine]:
        return self.program.local_routines

    def get_request_routine(self, name: str) -> RequestRoutine:
        return self.program.request_routines[name]

    def get_all_request_routines(self) -> Dict[str, RequestRoutine]:
        return self.program.request_routines

    @property
    def pid(self) -> int:
        return self.prog_instance.pid  # type: ignore

    @property
    def program(self) -> QoalaProgram:
        return self.prog_instance.program

    @property
    def inputs(self) -> ProgramInput:
        return self.prog_instance.inputs

    @property
    def host_mem(self) -> HostMemory:
        return self.prog_memory.host_mem

    @property
    def qnos_mem(self) -> QnosMemory:
        return self.prog_memory.qnos_mem

    @property
    def shared_mem(self) -> SharedMemory:
        return self.prog_memory.shared_mem



# from __future__ import annotations

# import os
# from dataclasses import dataclass
# from typing import Dict, List

# import netsquid as ns
# from netqasm.lang.instr.flavour import Flavour, NVFlavour, VanillaFlavour

# from qoala.lang.ehi import UnitModule
# from qoala.lang.parse import QoalaParser
# from qoala.lang.program import QoalaProgram
# from qoala.runtime.config import (
#     LatenciesConfig,
#     NVQDeviceConfig,
#     ProcNodeConfig,
#     ProcNodeNetworkConfig,
#     TopologyConfig,
# )
# from qoala.runtime.environment import NetworkInfo
# from qoala.runtime.lhi import LhiTopology, LhiTopologyBuilder
# from qoala.runtime.program import BatchInfo, BatchResult, ProgramInput
# from qoala.runtime.schedule import TaskSchedule, TaskScheduleEntry
# from qoala.sim.build import build_network


# def create_network_info(names: List[str]) -> NetworkInfo:
#     env = NetworkInfo.with_nodes({i: name for i, name in enumerate(names)})
#     env.set_global_schedule([0, 1, 2])
#     env.set_timeslot_len(1e6)
#     return env


# def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
#     return ProcNodeConfig(
#         node_name=name,
#         node_id=id,
#         # topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
#         latencies=LatenciesConfig(
#             host_instr_time=500, host_peer_latency=100_000, qnos_instr_time=1000
#         ),
#         nv_config=NVQDeviceConfig.perfect_config(num_qubits),
#     )


# def load_program(path: str) -> QoalaProgram:
#     path = os.path.join(os.path.dirname(__file__), path)
#     with open(path) as file:
#         text = file.read()
#     return QoalaParser(text, flavour=NVFlavour()).parse()


# def create_batch(
#     program: QoalaProgram,
#     unit_module: UnitModule,
#     inputs: List[ProgramInput],
#     num_iterations: int,
# ) -> BatchInfo:
#     return BatchInfo(
#         program=program,
#         unit_module=unit_module,
#         inputs=inputs,
#         num_iterations=num_iterations,
#         deadline=0,
#     )


# @dataclass
# class PingPongResult:
#     alice_results: Dict[int, BatchResult]
#     bob_results: Dict[int, BatchResult]


# def run_pingpong(num_iterations: int) -> PingPongResult:
#     ns.sim_reset()

#     num_qubits = 3
#     network_info = create_network_info(names=["bob", "alice"])
#     alice_id = network_info.get_node_id("alice")
#     bob_id = network_info.get_node_id("bob")

#     alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits)
#     bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits)

#     network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
#         nodes=[alice_node_cfg, bob_node_cfg], link_duration=500_000
#     )
#     network = build_network(network_cfg, network_info)
#     alice_procnode = network.nodes["alice"]
#     bob_procnode = network.nodes["bob"]

#     alice_program = load_program("pingpong_nv_alice.iqoala")
#     alice_inputs = [ProgramInput({"bob_id": bob_id}) for _ in range(num_iterations)]

#     alice_unit_module = UnitModule.from_full_ehi(alice_procnode.memmgr.get_ehi())
#     alice_batch = create_batch(
#         alice_program, alice_unit_module, alice_inputs, num_iterations
#     )
#     alice_procnode.submit_batch(alice_batch)
#     alice_procnode.initialize_processes()
#     alice_tasks = alice_procnode.scheduler.get_tasks_to_schedule()
#     print("Alice tasks:")
#     print([str(t) for t in alice_tasks])
#     # alice_schedule = TaskSchedule.consecutive(alice_tasks)
#     alice_schedule = TaskSchedule(
#         [
#             TaskScheduleEntry(alice_tasks[0], timestamp=0),
#             TaskScheduleEntry(alice_tasks[1], timestamp=500),
#             TaskScheduleEntry(alice_tasks[2], timestamp=1_025_000),
#             TaskScheduleEntry(alice_tasks[3], timestamp=1_600_000),
#             TaskScheduleEntry(alice_tasks[4], timestamp=9_850_000, prev=alice_tasks[3]),
#             TaskScheduleEntry(alice_tasks[5], timestamp=10_200_000),
#             TaskScheduleEntry(
#                 alice_tasks[6], timestamp=17_700_000, prev=alice_tasks[5]
#             ),
#             TaskScheduleEntry(alice_tasks[7], timestamp=29_100_000),
#             TaskScheduleEntry(
#                 alice_tasks[8], timestamp=29_200_500, prev=alice_tasks[7]
#             ),
#             TaskScheduleEntry(alice_tasks[9], timestamp=31_220_000),
#             TaskScheduleEntry(
#                 alice_tasks[10], timestamp=34_228_000, prev=alice_tasks[9]
#             ),
#         ]
#     )
#     print("\nAlice schedule:")
#     print(alice_schedule)
#     alice_procnode.scheduler.upload_schedule(alice_schedule)

#     bob_program = load_program("pingpong_nv_bob.iqoala")
#     bob_inputs = [ProgramInput({"alice_id": alice_id}) for _ in range(num_iterations)]

#     bob_unit_module = UnitModule.from_full_ehi(bob_procnode.memmgr.get_ehi())
#     bob_batch = create_batch(bob_program, bob_unit_module, bob_inputs, num_iterations)
#     bob_procnode.submit_batch(bob_batch)
#     bob_procnode.initialize_processes()
#     bob_tasks = bob_procnode.scheduler.get_tasks_to_schedule()
#     print("\n\nBob tasks:")
#     print([str(t) for t in bob_tasks])
#     bob_schedule = TaskSchedule(
#         [
#             TaskScheduleEntry(bob_tasks[0], timestamp=0),
#             TaskScheduleEntry(bob_tasks[1], timestamp=25_000),
#             TaskScheduleEntry(bob_tasks[2], timestamp=1_900_000, prev=bob_tasks[1]),
#             TaskScheduleEntry(bob_tasks[3], timestamp=12_000_000),
#             TaskScheduleEntry(bob_tasks[4], timestamp=12_100_000, prev=bob_tasks[3]),
#             TaskScheduleEntry(bob_tasks[5], timestamp=13_200_000),
#             TaskScheduleEntry(bob_tasks[6], timestamp=15_700_000),
#             TaskScheduleEntry(bob_tasks[7], timestamp=16_800_000),
#             TaskScheduleEntry(bob_tasks[8], timestamp=27_900_000, prev=bob_tasks[7]),
#         ]
#     )
#     print("\nBob schedule:")
#     print(bob_schedule)
#     bob_procnode.scheduler.upload_schedule(bob_schedule)

#     network.start()
#     ns.sim_run()

#     alice_results = alice_procnode.scheduler.get_batch_results()
#     bob_results = bob_procnode.scheduler.get_batch_results()

#     return PingPongResult(alice_results, bob_results)


# def test_pingpong():
#     # LogManager.set_log_level("INFO")

#     def check(num_iterations):
#         ns.sim_reset()
#         result = run_pingpong(num_iterations=num_iterations)
#         assert len(result.alice_results) > 0
#         assert len(result.bob_results) > 0

#         alice_batch_results = result.alice_results
#         for _, batch_results in alice_batch_results.items():
#             program_results = batch_results.results
#             outcomes = [result.values["outcome"] for result in program_results]
#             assert all(outcome == 1 for outcome in outcomes)

#     check(1)


if __name__ == "__main__":
    # Needs a fix, see #50.
    # test_pingpong()
    pass



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Dict, List, Optional

import netsquid as ns

from qoala.lang.ehi import EhiNodeInfo, UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.environment import NetworkInfo
from qoala.runtime.program import ProgramInput, ProgramInstance
from qoala.runtime.schedule import TaskSchedule
from qoala.runtime.task import TaskCreator, TaskExecutionMode
from qoala.sim.build import build_network
from qoala.util.logging import LogManager


def create_network_info(names: List[str]) -> NetworkInfo:
    env = NetworkInfo.with_nodes({i: name for i, name in enumerate(names)})
    env.set_global_schedule([0, 1, 2])
    env.set_timeslot_len(1e6)
    return env


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(
            host_instr_time=500, host_peer_latency=1_000_000, qnos_instr_time=1000
        ),
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class SimpleBqcResult:
    client_result: Dict[str, int]
    server_result: Dict[str, int]


def instantiate(
    program: QoalaProgram,
    ehi: EhiNodeInfo,
    pid: int = 0,
    inputs: Optional[ProgramInput] = None,
) -> ProgramInstance:
    unit_module = UnitModule.from_full_ehi(ehi)

    if inputs is None:
        inputs = ProgramInput.empty()

    return ProgramInstance(
        pid,
        program,
        inputs,
        unit_module=unit_module,
        block_tasks=[],
    )


def run_bqc(alpha, beta, theta1, theta2) -> SimpleBqcResult:
    ns.sim_reset()

    num_qubits = 3
    network_info = create_network_info(names=["client", "server"])
    server_id = network_info.get_node_id("server")
    client_id = network_info.get_node_id("client")

    server_node_cfg = create_procnode_cfg("server", server_id, num_qubits)
    client_node_cfg = create_procnode_cfg("client", client_id, num_qubits)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[server_node_cfg, client_node_cfg], link_duration=1000
    )
    network = build_network(network_cfg, network_info)
    server_procnode = network.nodes["server"]
    client_procnode = network.nodes["client"]

    server_program = load_program("bqc_server.iqoala")
    server_input = ProgramInput({"client_id": client_id})
    server_instance = instantiate(
        server_program, server_procnode.local_ehi, 0, server_input
    )
    server_procnode.scheduler.submit_program_instance(server_instance)

    client_program = load_program("bqc_client.iqoala")
    client_input = ProgramInput(
        {
            "server_id": server_id,
            "alpha": alpha,
            "beta": beta,
            "theta1": theta1,
            "theta2": theta2,
        }
    )
    client_instance = instantiate(
        client_program, client_procnode.local_ehi, 0, client_input
    )
    client_procnode.scheduler.submit_program_instance(client_instance)

    task_creator = TaskCreator(mode=TaskExecutionMode.ROUTINE_ATOMIC)
    tasks_server = task_creator.from_program(
        server_program,
        0,
        server_procnode.local_ehi,
        server_procnode.network_ehi,
        client_id,
    )
    tasks_client = task_creator.from_program(
        client_program,
        0,
        client_procnode.local_ehi,
        client_procnode.network_ehi,
        server_id,
    )

    schedule_server = TaskSchedule.consecutive(tasks_server)
    schedule_client = TaskSchedule.consecutive(tasks_client)
    server_procnode.scheduler.upload_schedule(schedule_server)
    client_procnode.scheduler.upload_schedule(schedule_client)

    network.start()
    ns.sim_run()

    client_result = client_procnode.memmgr.get_process(0).result.values
    server_result = server_procnode.memmgr.get_process(0).result.values

    return SimpleBqcResult(client_result, server_result)


def test_bqc():
    # Effective computation: measure in Z the following state:
    # H Rz(beta) H Rz(alpha) |+>
    # m2 should be this outcome

    # angles are in multiples of pi/16

    LogManager.set_log_level("WARNING")

    def check(alpha, beta, theta1, theta2, expected, num_iterations):
        ns.sim_reset()

        for _ in range(num_iterations):
            bqc_result = run_bqc(alpha=alpha, beta=beta, theta1=theta1, theta2=theta2)
            assert bqc_result.server_result["m2"] == expected

    check(alpha=8, beta=8, theta1=0, theta2=0, expected=0, num_iterations=10)
    check(alpha=8, beta=24, theta1=0, theta2=0, expected=1, num_iterations=10)
    check(alpha=8, beta=8, theta1=13, theta2=27, expected=0, num_iterations=10)
    check(alpha=8, beta=24, theta1=2, theta2=22, expected=1, num_iterations=10)


if __name__ == "__main__":
    test_bqc()



from __future__ import annotations

from typing import Generator, List, Optional, Tuple

from pydynaa import EventExpression
from qoala.runtime.message import Message
from qoala.sim.host.csocket import ClassicalSocket
from qoala.sim.host.hostinterface import HostInterface
from qoala.util.tests import yield_from


class MockHostInterface(HostInterface):
    def __init__(self) -> None:
        self.remote: Optional[MockHostInterface] = None
        self.messages: List[Message] = []

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        assert self.remote is not None
        self.remote.messages.append(msg)

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        return self.messages.pop()
        yield  # to make it behave as a generator


def setup_alice_bob() -> Tuple[ClassicalSocket, ClassicalSocket]:
    alice = MockHostInterface()
    bob = MockHostInterface()
    alice.remote = bob
    bob.remote = alice

    return (ClassicalSocket(alice, "bob"), ClassicalSocket(bob, "alice"))


def test_send_str():
    alice, bob = setup_alice_bob()

    alice.send("hello")
    msg = yield_from(bob.recv())
    assert msg == "hello"


def test_send_int():
    alice, bob = setup_alice_bob()

    alice.send_int(3)
    msg = yield_from(bob.recv_int())
    assert msg == 3


def test_send_float():
    alice, bob = setup_alice_bob()

    alice.send_float(3.14)
    msg = yield_from(bob.recv_float())
    assert msg == 3.14


if __name__ == "__main__":
    test_send_str()



from dataclasses import dataclass
from typing import List


@dataclass(eq=True, frozen=True)
class MultiQubit:
    qubit_ids: List[int]

    def __hash__(self) -> int:
        return hash(tuple(self.qubit_ids))



import logging
from typing import Dict, Generator, List

from netsquid.components.component import Component, Port
from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.runtime.message import Message
from qoala.util.logging import LogManager


class PortListener(Protocol):
    def __init__(self, port: Port, signal_label: str) -> None:
        self._buffer: List[Message] = []
        self._port: Port = port
        self._signal_label = signal_label
        self.add_signal(signal_label)

    @property
    def buffer(self) -> List[Message]:
        return self._buffer

    def run(self) -> Generator[EventExpression, None, None]:
        while True:
            # Wait for an event saying that there is new input.
            yield self.await_port_input(self._port)

            counter = 0
            # Read all inputs and count them.
            while True:
                input = self._port.rx_input()
                if input is None:
                    break
                self._buffer += input.items
                counter += 1
            # If there are n inputs, there have been n events, but we yielded only
            # on one of them so far. "Flush" these n-1 additional events:
            while counter > 1:
                yield self.await_port_input(self._port)
                counter -= 1

            # Only after having yielded on all current events, we can schedule a
            # notification event, so that its reactor can handle all inputs at once.
            self.send_signal(self._signal_label)


class ComponentProtocol(Protocol):
    def __init__(self, name: str, comp: Component) -> None:
        super().__init__(name)
        self._listeners: Dict[str, PortListener] = {}
        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({comp.name})"
        )

    def add_listener(self, name, listener: PortListener) -> None:
        self._listeners[name] = listener

    def _receive_msg(
        self, listener_name: str, wake_up_signal: str
    ) -> Generator[EventExpression, None, Message]:
        listener = self._listeners[listener_name]
        if len(listener.buffer) == 0:
            yield self.await_signal(sender=listener, signal_label=wake_up_signal)
        return listener.buffer.pop(0)

    def _receive_msg_any_source(
        self, listener_names: List[str], wake_up_signals: List[str]
    ) -> Generator[EventExpression, None, Message]:
        # TODO rewrite two separate lists as function arguments

        # First check if there is any listener with messages in their buffer.
        for listener_name, wake_up_signal in zip(listener_names, wake_up_signals):
            listener = self._listeners[listener_name]
            if len(listener.buffer) != 0:
                return listener.buffer.pop(0)

        # Else, get an EventExpression for each listener.
        expressions: List[EventExpression] = []

        for listener_name, wake_up_signal in zip(listener_names, wake_up_signals):
            listener = self._listeners[listener_name]
            assert len(listener.buffer) == 0  # already checked this above
            ev_expr = self.await_signal(sender=listener, signal_label=wake_up_signal)
            expressions.append(ev_expr)

        # Create a union of all expressoins.
        assert len(expressions) > 0
        union = expressions[0]
        for i in range(1, len(expressions)):
            union = union | expressions[i]  # type: ignore

        # Yield until at least one of the listeners got a message.
        yield union

        # Count the messages in listener's buffers.
        msg_count = 0
        msg_to_return = None
        for listener_name, wake_up_signal in zip(listener_names, wake_up_signals):
            listener = self._listeners[listener_name]

            # Count the number of new messages.
            msg_count += len(listener.buffer)

            # Only the first buffered message we encounter is actually popped
            # and returned. The others messages will get returned in a new call
            # to `_receive_msg_any_source`.
            if len(listener.buffer) != 0 and msg_to_return is None:
                msg_to_return = listener.buffer.pop(0)

        # There *must* be at least one since we yielded.
        assert msg_count > 0
        # "Flush away" the events that were also in the union.
        # We already yielded on one (above), but need to yield on the rest.
        for _ in range(msg_count - 1):
            yield union

        assert msg_to_return is not None
        return msg_to_return

    def start(self) -> None:
        super().start()
        for listener in self._listeners.values():
            listener.start()

    def stop(self) -> None:
        for listener in self._listeners.values():
            listener.stop()
        super().stop()



from __future__ import annotations

from typing import Generator

import netsquid as ns
from netsquid.nodes import Node

from pydynaa import EventExpression
from qoala.runtime.environment import LocalEnvironment, NetworkInfo
from qoala.runtime.message import Message
from qoala.sim.netstack import NetstackComponent, NetstackInterface


class MockNetstackInterface(NetstackInterface):
    def __init__(self, comp: NetstackComponent, local_env: LocalEnvironment) -> None:
        super().__init__(comp, local_env, None, None)


def create_netstackcomp(num_other_nodes: int) -> NetstackComponent:
    node = Node(name="alice", ID=0)

    nodes = {id: f"node_{id}" for id in range(1, num_other_nodes + 1)}
    nodes[0] = "alice"
    env = NetworkInfo.with_nodes(nodes)

    return NetstackComponent(node, env)


def test_no_other_nodes():
    comp = create_netstackcomp(num_other_nodes=0)

    # should have host_{in|out}, qnos_{in|out}, qnos_mem_{in|out}, entdist_{in|out} ports
    assert len(comp.ports) == 8
    assert "host_in" in comp.ports
    assert "host_out" in comp.ports
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "qnos_mem_in" in comp.ports
    assert "qnos_mem_out" in comp.ports
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports


def test_one_other_node():
    comp = create_netstackcomp(num_other_nodes=1)

    # should have 2 host ports + 4 qnos ports + 2 entdist ports + 2 peer ports
    assert len(comp.ports) == 10
    assert "host_in" in comp.ports
    assert "host_out" in comp.ports
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "qnos_mem_in" in comp.ports
    assert "qnos_mem_out" in comp.ports
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports

    assert "peer_node_1_in" in comp.ports
    assert "peer_node_1_out" in comp.ports

    # Test properties
    assert comp.peer_in_port("node_1") == comp.ports["peer_node_1_in"]
    assert comp.peer_out_port("node_1") == comp.ports["peer_node_1_out"]


def test_many_other_nodes():
    comp = create_netstackcomp(num_other_nodes=5)

    # should have 2 host ports + 4 qnos ports + 2 entdist ports + 5 * 2 peer ports
    assert len(comp.ports) == 18
    assert "host_in" in comp.ports
    assert "host_out" in comp.ports
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "qnos_mem_in" in comp.ports
    assert "qnos_mem_out" in comp.ports
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports

    for i in range(1, 6):
        assert f"peer_node_{i}_in" in comp.ports
        assert f"peer_node_{i}_out" in comp.ports
        # Test properties
        assert comp.peer_in_port(f"node_{i}") == comp.ports[f"peer_node_{i}_in"]
        assert comp.peer_out_port(f"node_{i}") == comp.ports[f"peer_node_{i}_out"]


def test_connection():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    env = NetworkInfo.with_nodes({alice.ID: alice.name, bob.ID: bob.name})

    alice_comp = NetstackComponent(alice, env)
    bob_comp = NetstackComponent(bob, env)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message("hello"))

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello"

    alice_intf = AliceNetstackInterface(alice_comp, LocalEnvironment(env, alice.ID))
    bob_intf = BobNetstackInterface(bob_comp, LocalEnvironment(env, bob.ID))

    alice_intf.start()
    bob_intf.start()

    ns.sim_run()


def test_three_way_connection():
    ns.sim_reset()

    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    charlie = Node(name="charlie", ID=2)
    env = NetworkInfo.with_nodes(
        {alice.ID: alice.name, bob.ID: bob.name, charlie.ID: charlie.name}
    )

    alice_comp = NetstackComponent(alice, env)
    bob_comp = NetstackComponent(bob, env)
    charlie_comp = NetstackComponent(charlie, env)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))
    alice_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("alice"))
    bob_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("bob"))
    bob_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("bob"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message("hello bob"))
            self.send_peer_msg("charlie", Message("hello charlie"))

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello bob"

    class CharlieNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello charlie"

    alice_intf = AliceNetstackInterface(alice_comp, LocalEnvironment(env, alice.ID))
    bob_intf = BobNetstackInterface(bob_comp, LocalEnvironment(env, bob.ID))
    charlie_intf = CharlieNetstackInterface(
        charlie_comp, LocalEnvironment(env, charlie.ID)
    )

    alice_intf.start()
    bob_intf.start()
    charlie_intf.start()

    ns.sim_run()


if __name__ == "__main__":
    test_no_other_nodes()
    test_one_other_node()
    test_many_other_nodes()

    test_connection()
    test_three_way_connection()



from __future__ import annotations

from qoala.runtime.environment import NetworkInfo
from qoala.sim.procnodecomp import ProcNodeComponent


def create_procnodecomp(num_other_nodes: int) -> ProcNodeComponent:
    nodes = {id: f"node_{id}" for id in range(1, num_other_nodes + 1)}
    nodes[0] = "alice"
    env = NetworkInfo.with_nodes(nodes)

    return ProcNodeComponent(name="alice", qprocessor=None, network_info=env)


def test_no_other_nodes():
    comp = create_procnodecomp(num_other_nodes=0)

    # should not have 2 entdist ports
    assert len(comp.ports) == 2
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports


def test_one_other_node():
    comp = create_procnodecomp(num_other_nodes=1)

    # should have 2 host peer ports + 2 netstack peer ports + 2 entdist ports
    assert len(comp.ports) == 6
    assert "host_peer_node_1_in" in comp.ports
    assert "host_peer_node_1_out" in comp.ports
    assert "netstack_peer_node_1_in" in comp.ports
    assert "netstack_peer_node_1_out" in comp.ports
    assert "entdist_in" in comp.ports
    assert "entdist_out" in comp.ports

    # Test properties
    assert comp.host_peer_in_port("node_1") == comp.ports["host_peer_node_1_in"]
    assert comp.host_peer_out_port("node_1") == comp.ports["host_peer_node_1_out"]
    assert comp.netstack_peer_in_port("node_1") == comp.ports["netstack_peer_node_1_in"]
    assert (
        comp.netstack_peer_out_port("node_1") == comp.ports["netstack_peer_node_1_out"]
    )


def test_many_other_nodes():
    comp = create_procnodecomp(num_other_nodes=5)

    # should 5 * 4 peer ports + 2 entdist ports
    assert len(comp.ports) == 22

    for i in range(1, 6):
        assert f"host_peer_node_{i}_in" in comp.ports
        assert f"host_peer_node_{i}_out" in comp.ports
        assert f"netstack_peer_node_{i}_in" in comp.ports
        assert f"netstack_peer_node_{i}_out" in comp.ports
        # Test properties
        assert (
            comp.host_peer_in_port(f"node_{i}") == comp.ports[f"host_peer_node_{i}_in"]
        )
        assert (
            comp.host_peer_out_port(f"node_{i}")
            == comp.ports[f"host_peer_node_{i}_out"]
        )
        assert (
            comp.netstack_peer_in_port(f"node_{i}")
            == comp.ports[f"netstack_peer_node_{i}_in"]
        )
        assert (
            comp.netstack_peer_out_port(f"node_{i}")
            == comp.ports[f"netstack_peer_node_{i}_out"]
        )


if __name__ == "__main__":
    test_no_other_nodes()
    test_one_other_node()
    test_many_other_nodes()



from __future__ import annotations

from dataclasses import dataclass
from typing import Generator

from pydynaa import EventExpression
from qoala.runtime.environment import LocalEnvironment
from qoala.runtime.message import Message
from qoala.sim.componentprot import ComponentProtocol, PortListener
from qoala.sim.events import (
    EVENT_WAIT,
    SIGNAL_HOST_HOST_MSG,
    SIGNAL_NSTK_HOST_MSG,
    SIGNAL_QNOS_HOST_MSG,
)
from qoala.sim.host.hostcomp import HostComponent


@dataclass
class HostLatencies:
    host_instr_time: float = 0  # duration of classical Host instr execution
    host_peer_latency: float = 0  # processing time for messages from remote node

    @classmethod
    def all_zero(cls) -> HostLatencies:
        # NOTE: can also just use HostLatencies() which will default all values to 0
        # However, using this classmethod makes this behavior more explicit and clear.
        return HostLatencies(0, 0)


class HostInterface(ComponentProtocol):
    """NetSquid protocol representing a Host."""

    def __init__(
        self,
        comp: HostComponent,
        local_env: LocalEnvironment,
    ) -> None:
        """Host protocol constructor.

        :param comp: NetSquid component representing the Host
        """
        name = f"{comp.name}_protocol"
        super().__init__(name=name, comp=comp)
        self._comp = comp

        self._local_env = local_env

        self.add_listener(
            "qnos",
            PortListener(self._comp.qnos_in_port, SIGNAL_QNOS_HOST_MSG),
        )
        self.add_listener(
            "netstack",
            PortListener(self._comp.netstack_in_port, SIGNAL_NSTK_HOST_MSG),
        )
        for peer in self._local_env.get_all_other_node_names():
            self.add_listener(
                f"peer_{peer}",
                PortListener(
                    self._comp.peer_in_port(peer), f"{SIGNAL_HOST_HOST_MSG}_{peer}"
                ),
            )

    def send_qnos_msg(self, msg: Message) -> None:
        self._comp.qnos_out_port.tx_output(msg)

    def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
        return (yield from self._receive_msg("qnos", SIGNAL_QNOS_HOST_MSG))

    def send_netstack_msg(self, msg: Message) -> None:
        self._comp.netstack_out_port.tx_output(msg)

    def receive_netstack_msg(self) -> Generator[EventExpression, None, Message]:
        return (yield from self._receive_msg("netstack", SIGNAL_NSTK_HOST_MSG))

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self._comp.peer_out_port(peer).tx_output(msg)

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        return (
            yield from self._receive_msg(
                f"peer_{peer}", f"{SIGNAL_HOST_HOST_MSG}_{peer}"
            )
        )

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr



from __future__ import annotations

from typing import List, Set, Tuple

import pytest

from qoala.lang.ehi import EhiBuilder, UnitModule
from qoala.lang.program import ProgramMeta, QoalaProgram
from qoala.runtime.memory import ProgramMemory
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.sim.memmgr import AllocError, MemoryManager
from qoala.sim.process import QoalaProcess
from qoala.sim.qdevice import QDevice


class MockQDevice(QDevice):
    def __init__(self, typ: str = "star") -> None:
        self._typ = typ

    def get_all_qubit_ids(self) -> Set[int]:
        return {0, 1}

    def get_comm_qubit_ids(self) -> Set[int]:
        if self._typ == "star":
            return {0}
        else:
            assert self._typ == "uniform"
            return {0, 1}

    def get_non_comm_qubit_ids(self) -> Set[int]:
        if self._typ == "star":
            return {1}
        else:
            assert self._typ == "uniform"
            return {}

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        pass


def create_process(pid: int, unit_module: UnitModule) -> QoalaProcess:
    program = QoalaProgram(
        blocks=[],
        local_routines={},
        meta=ProgramMeta.empty("prog"),
    )
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=ProgramInput({}),
        unit_module=unit_module,
        block_tasks=[],
    )
    mem = ProgramMemory(pid=pid)

    process = QoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        result=ProgramResult(values={}),
    )
    return process


def create_unit_module_star() -> UnitModule:
    ehi = EhiBuilder.perfect_star(
        num_qubits=2,
        flavour=None,
        comm_instructions=[],
        comm_duration=0,
        mem_instructions=[],
        mem_duration=0,
        two_instructions=[],
        two_duration=0,
    )
    return UnitModule.from_full_ehi(ehi)


def create_unit_module_uniform() -> UnitModule:
    ehi = EhiBuilder.perfect_uniform(
        num_qubits=2,
        flavour=None,
        single_instructions=[],
        single_duration=0,
        two_instructions=[],
        two_duration=0,
    )
    return UnitModule.from_full_ehi(ehi)


def setup_manager(typ: str = "star") -> Tuple[int, MemoryManager]:
    if typ == "star":
        um = create_unit_module_star()
        qdevice = MockQDevice()
        mgr = MemoryManager("alice", qdevice)
    else:
        assert typ == "uniform"
        um = create_unit_module_uniform()
        qdevice = MockQDevice(typ="uniform")
        mgr = MemoryManager("alice", qdevice)

    process = create_process(0, um)
    mgr.add_process(process)
    pid = process.pid

    return pid, mgr


def setup_manager_multiple_processes(
    num_processes: int, typ: str = "star"
) -> Tuple[List[int], MemoryManager]:
    if typ == "star":
        um = create_unit_module_star()
        qdevice = MockQDevice()
        mgr = MemoryManager("alice", qdevice)
    else:
        assert typ == "uniform"
        um = create_unit_module_uniform()
        qdevice = MockQDevice(typ="uniform")
        mgr = MemoryManager("alice", qdevice)

    pids: List[int] = []

    for i in range(num_processes):
        process = create_process(i, um)
        mgr.add_process(process)
        pids.append(process.pid)

    return pids, mgr


def test_alloc_free_0():
    pid, mgr = setup_manager()

    assert mgr.phys_id_for(pid, 0) is None
    assert mgr.phys_id_for(pid, 1) is None
    assert mgr.virt_id_for(pid, 0) is None
    assert mgr.virt_id_for(pid, 1) is None

    mgr.allocate(pid, 0)
    assert mgr.phys_id_for(pid, 0) == 0
    assert mgr.phys_id_for(pid, 1) is None
    assert mgr.virt_id_for(pid, 0) == 0
    assert mgr.virt_id_for(pid, 1) is None

    mgr.free(pid, 0)
    assert mgr.phys_id_for(pid, 0) is None
    assert mgr.phys_id_for(pid, 1) is None
    assert mgr.virt_id_for(pid, 0) is None
    assert mgr.virt_id_for(pid, 1) is None


def test_alloc_free_0_1():
    pid, mgr = setup_manager()

    assert mgr.phys_id_for(pid, 0) is None
    assert mgr.phys_id_for(pid, 1) is None
    assert mgr.virt_id_for(pid, 0) is None
    assert mgr.virt_id_for(pid, 1) is None

    mgr.allocate(pid, 0)
    mgr.allocate(pid, 1)
    assert mgr.phys_id_for(pid, 0) == 0
    assert mgr.phys_id_for(pid, 1) == 1
    assert mgr.virt_id_for(pid, 0) == 0
    assert mgr.virt_id_for(pid, 1) == 1

    mgr.free(pid, 0)
    assert mgr.phys_id_for(pid, 0) is None
    assert mgr.phys_id_for(pid, 1) == 1
    assert mgr.virt_id_for(pid, 0) is None
    assert mgr.virt_id_for(pid, 1) == 1


def test_alloc_non_existing():
    pid, mgr = setup_manager()

    with pytest.raises(AllocError):
        mgr.allocate(pid, 2)


def test_alloc_already_allocated():
    pid, mgr = setup_manager()

    with pytest.raises(AllocError):
        mgr.allocate(pid, 1)
        mgr.allocate(pid, 1)


def test_free_alreay_freed():
    pid, mgr = setup_manager()

    with pytest.raises(AllocError):
        mgr.free(pid, 0)


def test_get_unmapped_qubit():
    pid, mgr = setup_manager()

    assert mgr.get_unmapped_non_comm_qubit(pid) == 1
    mgr.allocate(pid, 0)
    assert mgr.get_unmapped_non_comm_qubit(pid) == 1
    mgr.allocate(pid, 1)
    with pytest.raises(AllocError):
        mgr.get_unmapped_non_comm_qubit(pid)
    mgr.free(pid, 1)
    assert mgr.get_unmapped_non_comm_qubit(pid) == 1


def test_alloc_multiple_processes():
    [pid0, pid1], mgr = setup_manager_multiple_processes(2)

    assert mgr.phys_id_for(pid0, 0) is None
    assert mgr.phys_id_for(pid0, 1) is None
    assert mgr.virt_id_for(pid0, 0) is None
    assert mgr.virt_id_for(pid0, 1) is None
    assert mgr.phys_id_for(pid1, 0) is None
    assert mgr.phys_id_for(pid1, 1) is None
    assert mgr.virt_id_for(pid1, 0) is None
    assert mgr.virt_id_for(pid1, 1) is None

    mgr.allocate(pid0, 0)
    # Should allocate phys ID 0 for virt ID 0 of pid0
    assert mgr.phys_id_for(pid0, 0) == 0
    assert mgr.virt_id_for(pid0, 0) == 0

    with pytest.raises(AllocError):
        # Should try to allocate phys ID 0, but it's not available
        mgr.allocate(pid1, 0)

    mgr.allocate(pid1, 1)
    # Should have mapping:
    #   phys ID 0 : (pid0, virt0)
    #   phys ID 1 : (pid1, virt1)
    assert mgr.phys_id_for(pid0, 0) == 0
    assert mgr.phys_id_for(pid0, 1) is None
    assert mgr.virt_id_for(pid0, 0) == 0
    assert mgr.virt_id_for(pid0, 1) is None
    assert mgr.phys_id_for(pid1, 0) is None
    assert mgr.phys_id_for(pid1, 1) == 1
    assert mgr.virt_id_for(pid1, 0) is None
    assert mgr.virt_id_for(pid1, 1) == 1


def test_alloc_multiple_processes_same_virt_id():
    [pid0, pid1], mgr = setup_manager_multiple_processes(2, typ="uniform")

    mgr.allocate(pid0, 0)
    # Should allocate phys ID 0 for virt ID 0 of pid0
    assert mgr.phys_id_for(pid0, 0) == 0
    assert mgr.virt_id_for(pid0, 0) == 0

    # Should allocate phys ID 1 for virt ID 0 of pid1
    mgr.allocate(pid1, 0)
    assert mgr.phys_id_for(pid1, 0) == 1
    assert mgr.virt_id_for(pid1, 1) == 0

    with pytest.raises(AllocError):
        mgr.allocate(pid1, 1)


if __name__ == "__main__":
    test_alloc_free_0()
    test_alloc_free_0_1()
    test_alloc_non_existing()
    test_alloc_already_allocated()
    test_free_alreay_freed()
    test_get_unmapped_qubit()
    test_alloc_multiple_processes()
    test_alloc_multiple_processes_same_virt_id()



import os

from netqasm.lang.instr import core

from qoala.lang.hostlang import BasicBlockType
from qoala.lang.parse import QoalaParser
from qoala.runtime.environment import NetworkInfo
from qoala.runtime.lhi import (
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiProcNodeInfo,
    LhiTopologyBuilder,
)
from qoala.runtime.task import BlockTask, TaskCreator, TaskExecutionMode
from qoala.sim.build import build_network_from_lhi
from qoala.sim.network import ProcNodeNetwork


def relative_path(path: str) -> str:
    return os.path.join(os.getcwd(), os.path.dirname(__file__), path)


CL = BasicBlockType.CL
CC = BasicBlockType.CC
QL = BasicBlockType.QL
QC = BasicBlockType.QC


def setup_network() -> ProcNodeNetwork:
    topology = LhiTopologyBuilder.perfect_uniform_default_gates(num_qubits=3)
    latencies = LhiLatencies(
        host_instr_time=1000, qnos_instr_time=2000, host_peer_latency=3000
    )
    link_info = LhiLinkInfo.perfect(duration=20_000)

    alice_lhi = LhiProcNodeInfo(
        name="alice", id=0, topology=topology, latencies=latencies
    )
    network_lhi = LhiNetworkInfo.fully_connected([0, 1], link_info)
    network_info = NetworkInfo.with_nodes({0: "alice", 1: "bob"})
    bob_lhi = LhiProcNodeInfo(name="bob", id=1, topology=topology, latencies=latencies)
    return build_network_from_lhi([alice_lhi, bob_lhi], network_info, network_lhi)


def test_from_program_1():
    path = relative_path("integration/bqc/vbqc_client.iqoala")
    with open(path) as file:
        text = file.read()
    program = QoalaParser(text).parse()

    creator = TaskCreator(TaskExecutionMode.ROUTINE_ATOMIC)
    pid = 3
    tasks = creator.from_program(program, pid)

    assert tasks == [
        BlockTask(pid, "b0", CL),
        BlockTask(pid, "b1", QC),
        BlockTask(pid, "b2", QL),
        BlockTask(pid, "b3", QC),
        BlockTask(pid, "b4", QL),
        BlockTask(pid, "b5", CL),
        BlockTask(pid, "b6", CC),
        BlockTask(pid, "b7", CL),
        BlockTask(pid, "b8", CC),
        BlockTask(pid, "b9", CL),
    ]


def test_from_program_2():
    network = setup_network()
    alice = network.nodes["alice"]

    path = relative_path("test_scheduling_alice.iqoala")
    with open(path) as file:
        text = file.read()
    program = QoalaParser(text).parse()

    creator = TaskCreator(TaskExecutionMode.ROUTINE_ATOMIC)
    pid = 3
    tasks = creator.from_program(program, pid, alice.local_ehi, alice.network_ehi)

    cpu_time = alice.local_ehi.latencies.host_instr_time
    recv_time = alice.local_ehi.latencies.host_peer_latency
    qpu_time = alice.local_ehi.latencies.qnos_instr_time
    meas_time = alice.local_ehi.find_single_gate(0, core.MeasInstruction).duration
    epr_time = alice.network_ehi.get_link(0, 1).duration

    assert tasks == [
        BlockTask(pid, "blk_host0", CL, 2 * cpu_time),
        BlockTask(pid, "blk_host1", CL, 1 * cpu_time),
        BlockTask(pid, "blk_host2", CL, 1 * cpu_time),
        BlockTask(pid, "blk_prep_cc", CL, 2 * cpu_time),
        BlockTask(pid, "blk_send", CL, 1 * cpu_time),
        BlockTask(pid, "blk_recv", CC, 1 * recv_time),
        BlockTask(pid, "blk_add_one", QL, 5 * qpu_time),
        BlockTask(pid, "blk_epr_md_1", QC, 1 * epr_time),
        BlockTask(pid, "blk_epr_md_2", QC, 2 * epr_time),
        BlockTask(pid, "blk_epr_ck_1", QC, 1 * epr_time),
        BlockTask(pid, "blk_meas_q0", QL, 3 * qpu_time + 1 * meas_time),
        BlockTask(pid, "blk_epr_ck_2", QC, 2 * epr_time),
        BlockTask(pid, "blk_meas_q0_q1", QL, 6 * qpu_time + 2 * meas_time),
    ]


if __name__ == "__main__":
    test_from_program_1()
    test_from_program_2()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List

from qoala.lang.ehi import UnitModule
from qoala.lang.program import QoalaProgram
from qoala.runtime.schedule import BlockTask


@dataclass
class ProgramInput:
    values: Dict[str, Any]

    @classmethod
    def empty(cls) -> ProgramInput:
        return ProgramInput({})


@dataclass
class ProgramResult:
    values: Dict[str, Any]


@dataclass
class BatchInfo:
    """Description of a batch of program instances that should be executed."""

    program: QoalaProgram
    unit_module: UnitModule
    inputs: List[ProgramInput]  # dict of inputs for each iteration
    num_iterations: int
    deadline: float


@dataclass
class ProgramInstance:
    """A program instantiated with Program Inputs and a Unit Module"""

    pid: int
    program: QoalaProgram
    inputs: ProgramInput
    unit_module: UnitModule
    block_tasks: List[BlockTask]


@dataclass
class ProgramBatch:
    batch_id: int
    info: BatchInfo
    instances: List[ProgramInstance]


@dataclass
class BatchResult:
    batch_id: int
    results: List[ProgramResult]



import pytest

from qoala.runtime.config import (
    LatenciesConfig,
    LinkBetweenNodesConfig,
    LinkConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.environment import NetworkInfo
from qoala.sim.build import build_network


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(qnos_instr_time=1000),
    )


def test_perfect_links():
    server_id = 0
    client_id = 1
    network_info = NetworkInfo.with_nodes({server_id: "server", client_id: "client"})

    num_qubits = 1
    server_node_cfg = create_procnode_cfg("server", server_id, num_qubits)
    client_node_cfg = create_procnode_cfg("client", client_id, num_qubits)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[server_node_cfg, client_node_cfg], link_duration=1000
    )
    network = build_network(network_cfg, network_info)

    server = network.nodes["server"]
    link_info = server.network_ehi.get_link(server_id, client_id)
    assert link_info.duration == 1000
    assert link_info.fidelity == 1.0


def test_depolarise_links():
    server_id = 0
    client_id = 1
    network_info = NetworkInfo.with_nodes({server_id: "server", client_id: "client"})

    num_qubits = 1
    server_node_cfg = create_procnode_cfg("server", server_id, num_qubits)
    client_node_cfg = create_procnode_cfg("client", client_id, num_qubits)

    link_fidelity = 0.8
    link_cfg = LinkConfig.depolarise_config(fidelity=link_fidelity, state_delay=1000)
    link_between_cfg = LinkBetweenNodesConfig(
        node_id1=server_id, node_id2=client_id, link_config=link_cfg
    )
    network_cfg = ProcNodeNetworkConfig(
        nodes=[server_node_cfg, client_node_cfg], links=[link_between_cfg]
    )
    network = build_network(network_cfg, network_info)

    server = network.nodes["server"]
    link_info = server.network_ehi.get_link(server_id, client_id)
    assert link_info.duration == 1000
    assert link_info.fidelity == pytest.approx(0.8)


if __name__ == "__main__":
    test_perfect_links()
    test_depolarise_links()
